#  _________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright (c) 2014 Sandia Corporation.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  This software is distributed under the BSD License.
#  _________________________________________________________________________

### Ideas
# - Should be easy to warm start the benders script
#   from a history file, so one wouldn't need to start
#   from scratch

### SERIOUS TODO:
# - The ASL interface always returns duals. When the problem is a MIP,
#   they are always zero. We need to make sure that the subproblems
#   are not still MIPs after fixing the first-stage constraints;
#   otherwise, the benders cuts are invalid (because the duals are
#   artificially zero).

### TODOs:
# - feasibility cuts

import os
import sys
import logging
import time
import itertools

try:
    from collections import OrderedDict
except ImportError:                         #pragma:nocover
    from ordereddict import OrderedDict

from pyomo.opt import (SolverFactory,
                       TerminationCondition,
                       undefined)
from pyomo.core import (value, minimize, Set,
                        Objective, SOSConstraint,
                        Constraint, Var, RangeSet,
                        Expression, Suffix, Reals, Param)
from pyomo.core.base.constraint import _GeneralConstraintData
from pyomo.core.beta.list_objects import XConstraintList
from pyomo.pysp.util.configured_object import PySPConfiguredObject
from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_register_common_option,
                                    safe_register_unique_option,
                                    safe_declare_common_option,
                                    safe_declare_unique_option,
                                    _domain_percent,
                                    _domain_nonnegative,
                                    _domain_positive_integer,
                                    _domain_must_be_str,
                                    _domain_unit_interval,
                                    _domain_tuple_of_str,
                                    _domain_tuple_of_str_or_dict)
from pyomo.pysp.util.misc import (parse_command_line,
                                  launch_command)
from pyomo.pysp.scenariotree.manager import \
    (InvocationType,
     ScenarioTreeManager,
     ScenarioTreeManagerFactory)
from pyomo.pysp.scenariotree.manager_solver import \
    ScenarioTreeManagerSolverFactory
from pyomo.pysp.phutils import find_active_objective
from pyomo.pysp.ef import create_ef_instance
from pyomo.pysp.solvers.spsolver import (SPSolver,
                                         SPSolverResults,
                                         SPSolverFactory)

from six.moves import xrange

thisfile = os.path.abspath(__file__)
thisfile.replace(".pyc","").replace(".py","")

logger = logging.getLogger('pyomo.pysp')

_benders_group_label = "Benders Options"

def EXTERNAL_deactivate_rootnode_costs(manager,
                                       scenario):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios
    rootnode = manager.scenario_tree.findRootNode()
    scenario._instance.find_component(
        "PYSP_BENDERS_NODE_COST_TERM_"+rootnode.name).set_value(0.0)
    if manager.preprocessor is not None:
        manager.preprocessor.objective_updated[scenario.name] = True

def EXTERNAL_activate_rootnode_costs(manager,
                                     scenario):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios
    rootnode = manager.scenario_tree.findRootNode()
    nodecost_var = instance.find_component(
        rootnode._cost_variable[0])[rootnode._cost_variable[1]]
    scenario._instance.find_component(
        "PYSP_BENDERS_NODE_COST_TERM_"+rootnode.name).set_value(nodecost_var)
    if manager.preprocessor is not None:
        manager.preprocessor.objective_updated[scenario.name] = True

def EXTERNAL_activate_fix_constraints(manager,
                                      scenario):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios
    fix_constraint = scenario._instance.find_component(
        "PYSP_BENDERS_FIX_XHAT_CONSTRAINT")
    fix_constraint.activate()
    if manager.preprocessor is not None:
        preprocess_constraints_list = \
            manager.preprocessor.constraints_added_list[scenario.name]
        for constraint_data in fix_constraint.values():
            preprocess_constraints_list.append(constraint_data)

def EXTERNAL_deactivate_fix_constraints(manager,
                                        scenario):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios
    fix_constraint = scenario._instance.find_component(
        "PYSP_BENDERS_FIX_XHAT_CONSTRAINT")
    fix_constraint.deactivate()
    # no need to flag the preprocessor
    if manager.preprocessor is not None:
        preprocess_constraints_list = \
            manager.preprocessor.constraints_removed_list[scenario.name]
        for constraint_data in fix_constraint.values():
            preprocess_constraints_list.append(constraint_data)

def EXTERNAL_cleanup_from_benders(manager,
                                  scenario):

    instance = scenario._instance

    # restore node cost expressions
    for node in scenario.node_list:
        cost_term_name = "PYSP_BENDERS_NODE_COST_TERM_"+node.name
        assert hasattr(instance, cost_term_name)
        instance.del_component(cost_term_name)
    assert hasattr(scenario, "_instance_cost_expression_old")
    scenario._instance_cost_expression.set_value(
        scenario._instance_cost_expression_old)
    del scenario._instance_cost_expression_old

    if hasattr(scenario, "_remove_dual_at_benders_cleanup"):
        assert scenario._remove_dual_at_benders_cleanup
        del scenario._remove_dual_at_benders_cleanup
        if manager.scenario_tree.contains_bundles():
            found = 0
            for scenario_bundle in manager.scenario_tree.bundles:
                if scenario.name in scenario_bundle.scenario_names:
                    found += 1
                    bundle_instance = \
                        manager._bundle_binding_instance_map[scenario_bundle.name]
                    bundle_instance.del_component("dual")
            assert found == 1
        else:
            instance.del_component("dual")

    # restore cached domains
    scenario_bySymbol = instance._ScenarioTreeSymbolMap.bySymbol
    assert hasattr(instance, "PYSP_BENDERS_CACHED_DOMAINS")
    cached_domains = instance.PYSP_BENDERS_CACHED_DOMAINS
    for variable_id, domain, varbounds in cached_domains:
        vardata = scenario_bySymbol[variable_id]
        vardata.domain = domain
        vardata.setlb(varbounds[0])
        vardata.setub(varbounds[1])
    del instance.PYSP_BENDERS_CACHED_DOMAINS

    # remove fixing components
    nodal_index_set_name = "PYSP_BENDERS_FIX_XHAT_INDEX"
    assert hasattr(instance, nodal_index_set_name)
    instance.del_component(nodal_index_set_name)

    fix_param_name = "PYSP_BENDERS_FIX_XHAT_VALUE"
    assert hasattr(instance, fix_param_name)
    instance.del_component(fix_param_name)

    fix_constraint_name = "PYSP_BENDERS_FIX_XHAT_CONSTRAINT"
    assert hasattr(instance, fix_constraint_name)
    instance.del_component(fix_constraint_name)

    # The objective has changed so flag this if
    # necessary. Might as well flag the constraints as well
    # for safe measure
    if manager.preprocessor is not None:
        manager.preprocessor.objective_updated[scenario.name] = True

def EXTERNAL_initialize_for_benders(manager,
                                    scenario):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios

    rootnode = manager.scenario_tree.findRootNode()
    leafstage = scenario._leaf_node.stage
    instance = scenario._instance

    # disaggregate the objective into stage costs
    cost_terms = 0.0
    for node in scenario.node_list:
        stagecost_var = instance.find_component(
            node._cost_variable[0])[node._cost_variable[1]]
        cost_term_name = "PYSP_BENDERS_NODE_COST_TERM_"+node.name
        assert not hasattr(instance, cost_term_name)
        instance.add_component(cost_term_name,
                               Expression(initialize=stagecost_var))
        cost_terms += instance.find_component(cost_term_name)
    assert not hasattr(scenario, "_instance_cost_expression_old")
    scenario._instance_cost_expression_old = \
        scenario._instance_cost_expression.expr
    # we modify this component in place so that any bundle objectives
    # get updated automatically
    scenario._instance_cost_expression.set_value(cost_terms)

    # TODO: Remove first stage constraints?
    # NOTE: If any constraints are misidentified as
    #       first-stage (e.g., because they include only
    #       first-stage variables but have data that
    #       changes), then these constraints would not be
    #       accounted for in the overall problem if we
    #       deactivate them here, so it is better leave them
    #       all active by default.
    if manager.scenario_tree.contains_bundles():
        found = 0
        for scenario_bundle in manager.scenario_tree.bundles:
            if scenario.name in scenario_bundle.scenario_names:
                found += 1
                bundle_instance = \
                    manager._bundle_binding_instance_map[scenario_bundle.name]
                if not hasattr(bundle_instance,"dual"):
                    scenario._remove_dual_at_benders_cleanup = True
                    bundle_instance.dual = Suffix(direction=Suffix.IMPORT)
                else:
                    if isinstance(bundle_instance.dual, Suffix):
                        if not bundle_instance.dual.import_enabled():
                            print("Modifying existing dual component to import "
                                  "suffix data from solver.")
                            bundle_instance.dual.set_direction(Suffix.IMPORT_EXPORT)
                    else:
                        raise TypeError(
                            "Object with name 'dual' was found on model that "
                            "is not of type 'Suffix'. The object must be renamed "
                            "in order to use the benders algorithm.")
        assert found == 1
    else:
        if not hasattr(instance,"dual"):
            scenario._remove_dual_at_benders_cleanup = True
            instance.dual = Suffix(direction=Suffix.IMPORT)
        else:
            if isinstance(instance.dual, Suffix):
                if not instance.dual.import_enabled():
                    print("Modifying existing dual component to import "
                          "suffix data from solver.")
                    instance.dual.set_direction(Suffix.IMPORT_EXPORT)
            else:
                raise TypeError(
                    "Object with name 'dual' was found on model that "
                    "is not of type 'Suffix'. The object must be renamed "
                    "in order to use the benders algorithm.")

    # Relax all first-stage variables to be continuous, cache
    # their original bounds an domains on the instance so we
    # can restore them
    scenario_bySymbol = instance._ScenarioTreeSymbolMap.bySymbol
    assert not hasattr(instance, "PYSP_BENDERS_CACHED_DOMAINS")
    cached_domains = instance.PYSP_BENDERS_CACHED_DOMAINS = []
    # GH: Question: Is it possible that there are "derived"
    #               variables that are only "derived"
    #               when additional integrality conditions
    #               are placed on them? If so, they would
    #               need to be classified as "standard" in
    #               order to run benders. That is, unless we
    #               include all node variables in the
    #               benders cuts and fixing constraints by
    #               default. For now, we do not.
    for variable_id in rootnode._variable_ids:
        vardata = scenario_bySymbol[variable_id]
        # derived variables might be Expression objects
        if not vardata.is_expression_type():
            tight_bounds = vardata.bounds
            domain = vardata.domain
            vardata.domain = Reals
            # Collect the var bounds after setting the domain
            # to Reals. We do this so we know if the Var bounds
            # are set by the user or come from the domain
            varbounds = vardata.bounds
            cached_domains.append((variable_id, domain, varbounds))
            vardata.setlb(tight_bounds[0])
            vardata.setub(tight_bounds[1])

    # create index sets for fixing components
    nodal_index_set_name = "PYSP_BENDERS_FIX_XHAT_INDEX"
    assert not hasattr(instance, nodal_index_set_name)
    nodal_index_set = Set(name=nodal_index_set_name,
                          ordered=True,
                          initialize=sorted(rootnode._standard_variable_ids))
    instance.add_component(nodal_index_set_name, nodal_index_set)

    fix_param_name = "PYSP_BENDERS_FIX_XHAT_VALUE"
    assert not hasattr(instance, fix_param_name)
    instance.add_component(fix_param_name,
                           Param(nodal_index_set, mutable=True, initialize=0.0))
    fix_param = instance.find_component(fix_param_name)

    fix_constraint_name = "PYSP_BENDERS_FIX_XHAT_CONSTRAINT"
    assert not hasattr(instance, fix_constraint_name)
    def fix_rule(m,variable_id):
        # NOTE: The ordering within the expression is important here; otherwise
        #       duals will be returned with the opposite sign, which affects how
        #       the cut expressions need to be generated
        return  scenario_bySymbol[variable_id] - fix_param[variable_id] == 0.0
    instance.add_component(fix_constraint_name,
                           Constraint(nodal_index_set, rule=fix_rule))
    instance.find_component(fix_constraint_name).deactivate()

    # These will flag the necessary preprocessor info
    EXTERNAL_deactivate_rootnode_costs(manager, scenario)
    EXTERNAL_activate_fix_constraints(manager, scenario)

def EXTERNAL_update_fix_constraints(manager,
                                    scenario,
                                    fix_values):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios
    instance = scenario._instance
    fix_param_name = "PYSP_BENDERS_FIX_XHAT_VALUE"
    fix_param = instance.find_component(fix_param_name)
    fix_param.store_values(fix_values)

    fix_constraint = scenario._instance.find_component(
        "PYSP_BENDERS_FIX_XHAT_CONSTRAINT")
    if manager.preprocessor is not None:
        preprocess_constraints_list = \
            manager.preprocessor.constraints_updated_list[scenario.name]
        for constraint_data in fix_constraint.values():
            preprocess_constraints_list.append(constraint_data)

def EXTERNAL_collect_cut_data(manager,
                              scenario):
    assert len(manager.scenario_tree.stages) == 2
    assert scenario in manager.scenario_tree.scenarios

    dual_suffix = None
    sum_probability_bundle = None
    if manager.scenario_tree.contains_bundles():
        found = 0
        for scenario_bundle in manager.scenario_tree.bundles:
            if scenario.name in scenario_bundle.scenario_names:
                found += 1
                dual_suffix = \
                    manager._bundle_binding_instance_map[scenario_bundle.name].dual
                sum_probability_bundle = scenario_bundle.probability
        assert found == 1

    else:
        dual_suffix  = scenario._instance.dual
        sum_probability_bundle = scenario.probability
    rootnode = manager.scenario_tree.findRootNode()
    scenario_results = {}
    scenario_results['SSC'] = scenario._stage_costs[scenario.leaf_node.stage.name]
    duals = scenario_results['duals'] = {}
    benders_fix_constraint = scenario._instance.find_component(
        "PYSP_BENDERS_FIX_XHAT_CONSTRAINT")
    if len(dual_suffix) == 0:
        raise RuntimeError("No duals were returned with the solution for "
                           "scenario %s. This might indicate a solve failure or "
                           "that there are discrete variables in the second-stage "
                           "problem." % (scenario.name))
    for variable_id in rootnode._standard_variable_ids:
        duals[variable_id] = dual_suffix[benders_fix_constraint[variable_id]] \
                             * sum_probability_bundle \
                             / scenario.probability
    return scenario_results

class BendersOptimalityCut(object):
    __slots__ = ("xhat", "ssc", "duals")
    def __init__(self, xhat, ssc, duals):
        self.xhat = xhat
        self.ssc = ssc
        self.duals = duals

class BendersAlgorithm(PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()
        safe_declare_common_option(options,
                                   "verbose")
        safe_declare_unique_option(
            options,
            "max_iterations",
            PySPConfigValue(
                100,
                domain=_domain_positive_integer,
                description=(
                    "The maximum number of iterations. Default is 100."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "percent_gap",
            PySPConfigValue(
                0.0001,
                domain=_domain_percent,
                description=(
                    "Percent optimality gap required for convergence. "
                    "Default is 0.0001%%."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "multicut_level",
            PySPConfigValue(
                1,
                domain=int,
                description=(
                    "The number of cut groups added to the "
                    "master benders problem each iteration. "
                    "Default is 1. A number less than 1 indicates "
                    "that the maximum value should be used, which "
                    "is one cut group for each scenario not included "
                    "in the master problem."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "optimality_gap_epsilon",
            PySPConfigValue(
                1e-10,
                domain=_domain_nonnegative,
                description=(
                    "The epsilon value used in the denominator of "
                    "the optimality gap calculation. Default is 1e-10."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_include_scenarios",
            PySPConfigValue(
                (),
                domain=_domain_tuple_of_str,
                description=(
                    "A list of names of scenarios that should be included "
                    "in the master problem. This option can be used multiple "
                    "times from the command line to specify more than one "
                    "scenario name."
                ),
                doc=None,
                visibility=0),
            ap_kwds={'action': 'append'},
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_disable_warmstart",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "Disable warm-start of the benders master "
                    "problem solves. Default is False."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_solver",
            PySPConfigValue(
                "cplex",
                domain=_domain_must_be_str,
                description=(
                    "Specify the solver with which to solve "
                    "the master benders problem. Default is cplex."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_solver_io",
            PySPConfigValue(
                None,
                domain=_domain_must_be_str,
                description=(
                    "The type of IO used to execute the master "
                    "solver.  Different solvers support different "
                    "types of IO, but the following are common "
                    "options: lp - generate LP files, nl - generate "
                    "NL files, mps - generate MPS files, python - "
                    "direct Python interface, os - generate OSiL XML files."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_mipgap",
            PySPConfigValue(
                None,
                domain=_domain_unit_interval,
                description=(
                    "Specifies the mipgap for the master benders solves."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_solver_options",
            PySPConfigValue(
                (),
                domain=_domain_tuple_of_str_or_dict,
                description=(
                    "Persistent solver options used when solving the master "
                    "benders problem. This option can be used multiple times from "
                    "the command line to specify more than one solver option."
                ),
                doc=None,
                visibility=0),
            ap_kwds={'action': 'append'},
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_output_solver_log",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "Output solver log during solves of the master problem."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_keep_solver_files",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "Retain temporary input and output files for master "
                    "benders solves."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)
        safe_declare_unique_option(
            options,
            "master_symbolic_solver_labels",
            PySPConfigValue(
                False,
                domain=bool,
                description=(
                    "When interfacing with the solver, use "
                    "symbol names derived from the model. For "
                    "example, \"my_special_variable[1_2_3]\" "
                    "instead of \"v1\". Useful for "
                    "debugging. When using the ASL interface "
                    "(--solver-io=nl), generates corresponding "
                    ".row (constraints) and .col (variables) "
                    "files. The ordering in these files provides "
                    "a mapping from ASL index to symbolic model "
                    "names."
                ),
                doc=None,
                visibility=0),
            ap_group=_benders_group_label)

        ScenarioTreeManagerSolverFactory.register_options(
            options,
            options_prefix="subproblem_",
            setup_argparse=False)

        return options

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        if self._manager is not None:
            self.cleanup_subproblems()
        if self._manager_solver is not None:
            self._manager_solver.close()
        self._manager = None
        self._manager_solver = None
        self._master_solver = None

    def __init__(self, manager, *args, **kwds):
        # TODO: Does this import need to be delayed because
        #       it is in a plugins subdirectory
        from pyomo.solvers.plugins.solvers.persistent_solver import \
            PersistentSolver

        self._manager = None
        self._manager_solver = None
        self._master_solver = None

        # The following attributes will be modified by the
        # solve() method. For users that are scripting, these
        # can be accessed after the solve() method returns.
        # They will be reset each time solve() is called.
        ############################################
        # history of master bounds
        self.master_bound_history = OrderedDict()
        # history of objectives
        self.objective_history = OrderedDict()
        # best objective
        self.incumbent_objective = None
        # incumbent first-stage solution with best objective
        self.xhat = None
        # |best_bound - best_objective| / (eps + |best_objective|)
        self.optimality_gap = None
        # no. of iterations completed
        self.iterations = None
        ############################################

        # The following attributes will be modified by the
        # build_master_problem() method. They will be
        # reset each time it is called. Additionally,
        # cut_pool will be appended with a new cut at
        # each iteration within the solve() method.
        self.master = None
        self.cut_pool = []
        self._num_first_stage_constraints = None

        super(BendersAlgorithm, self).__init__(*args, **kwds)

        if not isinstance(manager, ScenarioTreeManager):
            raise TypeError("BendersAlgorithm requires an instance of the "
                            "ScenarioTreeManager interface as the "
                            "first argument")
        if not manager.initialized:
            raise ValueError("BendersAlgorithm requires a scenario tree "
                             "manager that has been fully initialized")
        if len(manager.scenario_tree.stages) != 2:
            raise ValueError("BendersAlgorithm requires a two-stage scenario tree")

        self._manager = manager
        self._manager_solver = ScenarioTreeManagerSolverFactory(
            self._manager,
            self._options,
            options_prefix="subproblem_")

        self._master_solver = None
        # setup the master solver
        self._master_solver = SolverFactory(
            self.get_option("master_solver"),
            solver_io=self.get_option("master_solver_io"))
        if isinstance(self._master_solver, PersistentSolver):
            raise TypeError("BendersAlgorithm does not yet support "
                            "PersistentSolver types for the master problem")
        if len(self.get_option("master_solver_options")):
            if type(self.get_option("master_solver_options")) is tuple:
                self._master_solver.set_options(
                    "".join(self.get_option("master_solver_options")))
            else:
                self._master_solver.set_options(
                    self.get_option("master_solver_options"))
        if self.get_option("master_mipgap") is not None:
            self._master_solver.options.mipgap = \
                self.get_option("master_mipgap")

    def deactivate_rootnode_costs(self):
        self._manager.invoke_function(
            "EXTERNAL_deactivate_rootnode_costs",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            oneway_call=True)

    def activate_rootnode_costs(self):
        self._manager.invoke_function(
            "EXTERNAL_activate_rootnode_costs",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            oneway_call=True)

    def activate_fix_constraints(self):
        self._manager.invoke_function(
            "EXTERNAL_activate_fix_constraints",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            oneway_call=True)

    def deactivate_fix_constraints(self):
        self._manager.invoke_function(
            "EXTERNAL_deactivate_fix_constraints",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            oneway_call=True)

    def update_fix_constraints(self, fix_values):
        self._manager.invoke_function(
            "EXTERNAL_update_fix_constraints",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            function_args=(fix_values,),
            oneway_call=True)

    def collect_cut_data(self, async_call=False):
        return self._manager.invoke_function(
            "EXTERNAL_collect_cut_data",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            async_call=async_call)

    def initialize_subproblems(self):
        if self.get_option("verbose"):
            print("Initializing subproblems for benders")
        self._manager.invoke_function(
            "EXTERNAL_initialize_for_benders",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            oneway_call=True)

    def cleanup_subproblems(self):
        if self.get_option("verbose"):
            print("Cleaning up subproblems for benders")
        self._manager.invoke_function(
            "EXTERNAL_cleanup_from_benders",
            thisfile,
            invocation_type=InvocationType.PerScenario,
            oneway_call=True)

    def generate_cut(self,
                     xhat,
                     update_stages=(),
                     return_solve_results=False):
        """
        Generate a cut for the first-stage solution xhat by
        solving the subproblems. By default, only the stage
        costs and objective values are updated on the local
        scenario tree. Setting update_stages to a list of
        stage names or None (indicating all stages) can be
        used to control how much solution information is
        loaded for the variables on the scenario tree.
        """
        self.update_fix_constraints(xhat)
        solve_results = \
            self._manager_solver.solve_subproblems()

        cut_data = self.collect_cut_data()
        benders_cut = BendersOptimalityCut(
            xhat,
            dict((name, cut_data[name]['SSC']) for name in cut_data),
            dict((name, cut_data[name]['duals']) for name in cut_data))

        if return_solve_results:
            return benders_cut, solve_results
        else:
            return benders_cut

    #
    # Any of the methods above this point can be called without
    # constructing a master problem.
    #

    def build_master_problem(self, **kwds):
        """
        Build a master problem to add cuts to. This method
        must called prior to calling methods like solve().
        When the optional keyword include_scenarios is used,
        it overrides the list of names of scenarios (if any)
        to include on the master that were specified on the options
        object used to initialize this class.
        """

        # check whether to override the
        # "master_include_scenarios" option
        if "include_scenarios" not in kwds:
            include_scenarios = self.get_option("master_include_scenarios")
        else:
            include_scenarios = kwds.pop("include_scenarios")
        assert len(kwds) == 0

        scenario_tree = self._manager.scenario_tree
        rootnode = scenario_tree.findRootNode()
        firststage = rootnode.stage
        objective_sense = self._manager.objective_sense

        # construct master problem
        if (include_scenarios is None) or \
           (len(include_scenarios) == 0):
            master_scenario_tree = scenario_tree.make_compressed(
                [scenario_tree.scenarios[0].name],
                normalize=False)
        else:
            print("Number of scenarios included in Benders master problem: %s"
                  % (len(include_scenarios)))
            master_scenario_tree = scenario_tree.make_compressed(
                include_scenarios,
                normalize=False)

        master_rootnode = master_scenario_tree.findRootNode()
        master_firststage = master_scenario_tree.stages[0]
        assert master_firststage is master_rootnode.stage
        master_secondstage = master_scenario_tree.stages[1]
        assert len(master_scenario_tree.stages) == 2

        master_scenario_instances = \
            scenario_tree._scenario_instance_factory.\
            construct_instances_for_scenario_tree(master_scenario_tree)
        # with the scenario instances now available, link the
        # referenced objects directly into the scenario tree.
        master_scenario_tree.linkInInstances(master_scenario_instances,
                                             create_variable_ids=True)

        master = create_ef_instance(master_scenario_tree)
        if (include_scenarios is None) or \
           (len(include_scenarios) == 0):
            master._scenarios_included = set()
            remove_ssc = True
        else:
            master._scenarios_included = \
                set(include_scenarios)
            remove_ssc = False

        #
        # Deactivate second-stage constraints
        #

        # first count the binding constraints on the top-level ef model
        num_first_stage_constraints = len(list(itertools.chain(
            master.component_data_objects(SOSConstraint,
                                          active=True,
                                          descend_into=False),
            master.component_data_objects(Constraint,
                                          active=True,
                                          descend_into=False))))
        # now count the first-stage constraints on the scenario
        # instances included in the master ef
        for scenario in master_scenario_tree.scenarios:
            instance = scenario._instance
            for constraint_data in itertools.chain(
                    instance.component_data_objects(SOSConstraint,
                                                    active=True,
                                                    descend_into=True),
                    instance.component_data_objects(Constraint,
                                                    active=True,
                                                    descend_into=True)):
                # Note that it is possible that we are misidentifying
                # some constraints as belonging to the first stage. This
                # would be the case when no second-stage variables appear
                # in the expression but one or more rhs or first-stage variable
                # coefficients changes with scenarios.
                try:
                    node = scenario.constraintNode(constraint_data,
                                                   instance=instance)
                except RuntimeError:
                    # TODO: Adapt this to handle the multistage case
                    node = scenario.node_list[-1]
                # Not sure if we want to allow variables to not be declared on
                # some stage in the scenario tree.
                                               #assume_last_stage_if_missing=True)
                if node.stage is not master_firststage:
                    assert node.stage is master_secondstage
                    if remove_ssc:
                        constraint_data.deactivate()
                else:
                    num_first_stage_constraints += 1

        self._num_first_stage_constraints = num_first_stage_constraints
        # deactivate original objective
        find_active_objective(master, safety_checks=True).deactivate()

        # add cut variable(s)
        if self.get_option("multicut_level") < 1:
            print("Using maximum number of cut groups")
            cut_bundles = [[] for scenario in scenario_tree.scenarios]
        else:
            cut_bundles = [[] for i in xrange(self.get_option("multicut_level"))]

        # TODO: Allow users some control over these cut_bundles
        cnt = 0
        len_cut_bundles = len(cut_bundles)
        assert len_cut_bundles > 0
        for scenario in scenario_tree.scenarios:
            if scenario.name not in master._scenarios_included:
                cut_bundles[cnt % len_cut_bundles].append(scenario.name)
                cnt += 1
        nonempty_cut_bundles = []
        for bundle in cut_bundles:
            if len(bundle) > 0:
                nonempty_cut_bundles.append(bundle)
        if len(nonempty_cut_bundles) != len(cut_bundles):
            if self.get_option("multicut_level") >= 1:
                print("The number of cut groups indicated by the multicut_level "
                      "option was too large. Reducing from %s to %s."
                      % (len(cut_bundles), len(nonempty_cut_bundles)))

        alpha_varname = "PYSP_BENDERS_ALPHA_SSC"
        assert not hasattr(master, alpha_varname)
        master.add_component(alpha_varname,Var())
        master_alpha = master.find_component(alpha_varname)

        alpha_bundles_index_name = "PYSP_BENDERS_BUNDLE_ALPHA_SSC_INDEX"
        assert not hasattr(master, alpha_bundles_index_name)
        master.add_component(alpha_bundles_index_name,
                             RangeSet(0,len(nonempty_cut_bundles)-1))
        bundle_alpha_index = master.find_component(alpha_bundles_index_name)

        alpha_bundles_name = "PYSP_BENDERS_BUNDLE_ALPHA_SSC"
        assert not hasattr(master, alpha_bundles_name)
        master.add_component(alpha_bundles_name,
                             Var(bundle_alpha_index))
        bundle_alpha = master.find_component(alpha_bundles_name)

        cut_bundles_list_name = "PYSP_BENDERS_CUT_BUNDLES_SSC"
        assert not hasattr(master, cut_bundles_list_name)
        setattr(master, cut_bundles_list_name, nonempty_cut_bundles)
        alpha_cut_constraint_name = "PYSP_BUNDLE_AVERAGE_ALPHA_CUT_SSC"
        assert not hasattr(master, alpha_cut_constraint_name)
        if objective_sense == minimize:
            master.add_component(
                alpha_cut_constraint_name,
                Constraint(expr=master_alpha >= sum(bundle_alpha[i]
                                                    for i in bundle_alpha_index)))
        else:
            master.add_component(
                alpha_cut_constraint_name,
                Constraint(expr=master_alpha <= sum(bundle_alpha[i]
                                                    for i in bundle_alpha_index)))

        # add new objective
        if (include_scenarios is None) or \
           (len(include_scenarios) == 0):
            assert len(master_scenario_tree.scenarios) == 1
            master_cost_expr = master.find_component(
                master_scenario_tree.scenarios[0].name).find_component(
                    master_rootnode._cost_variable[0])\
                    [master_rootnode._cost_variable[1]]
        else:
            # NOTE: We include the first-stage cost expression for
            #       each of the scenarios included in the master with
            #       normalized probabilities. The second stage costs
            #       are included without normalizing probabilities so
            #       that they can simply be excluded from any cut expressions
            #       without having to re-normalize anything in the cuts
            master_cost_expr = 0.0
            normalization = sum(scenario.probability
                                for scenario in master_scenario_tree.scenarios)
            for scenario in master_scenario_tree.scenarios:
                scenario_rootnode = scenario.node_list[0]
                assert scenario_rootnode is master_rootnode
                rootnode_cost_expr = scenario._instance.find_component(
                    scenario_rootnode._cost_variable[0])\
                    [scenario_rootnode._cost_variable[1]]
                scenario_leafnode = scenario.node_list[1]
                leafnode_cost_expr = scenario._instance.find_component(
                    scenario_leafnode._cost_variable[0])\
                    [scenario_leafnode._cost_variable[1]]
                master_cost_expr += scenario.probability * \
                                    (rootnode_cost_expr / normalization + \
                                     leafnode_cost_expr)
        benders_objective_name = "PYSP_BENDERS_OBJECTIVE"
        assert not hasattr(master, benders_objective_name)
        master.add_component(
            benders_objective_name,
            Objective(expr=master_cost_expr + master_alpha,
                      sense=objective_sense))
        master_objective = master.find_component(
            benders_objective_name)
        cutlist_constraint_name = "PYSP_BENDERS_CUTS_SSC"
        assert not hasattr(master, cutlist_constraint_name)
        # I am using the XConstraintList prototype because
        # it is zero-based, meaning the index within self.cut_pool
        # (which stores the benders cuts objects) will correspond
        # directly with the index within this constraint.
        master.add_component(cutlist_constraint_name,
                             XConstraintList())

        self.master = master
        self.cut_pool = []

    def add_cut(self, benders_cut, ignore_cut_bundles=False):
        """
        Add the cut defined by the benders_cut object to the
        master problem. The optional keyword ignore_cut_bundles
        can be used generate the cut using the single master
        alpha cut variable rather than over the possibly many
        bundle cut groups.
        """

        if self.master is None:
            raise RuntimeError("The master problem has not been constructed."
                               "Call the build_master_problem() method to "
                               "construct it.")

        # for now, until someone figures out feasibility cuts
        assert benders_cut.__class__ is BendersOptimalityCut

        self.cut_pool.append(benders_cut)

        scenario_tree = self._manager.scenario_tree
        objective_sense = self._manager.objective_sense
        master = self.master
        rootnode = scenario_tree.findRootNode()
        master_variable = master.find_component(
            "MASTER_BLEND_VAR_"+str(rootnode.name))
        benders_cuts = master.find_component(
            "PYSP_BENDERS_CUTS_SSC")
        master_alpha = master.find_component(
            "PYSP_BENDERS_ALPHA_SSC")
        bundle_alpha = master.find_component(
            "PYSP_BENDERS_BUNDLE_ALPHA_SSC")

        xhat = benders_cut.xhat
        cut_expression = 0.0
        if not ignore_cut_bundles:
            for i, cut_scenarios in enumerate(
                    getattr(master, "PYSP_BENDERS_CUT_BUNDLES_SSC")):

                for scenario_name in cut_scenarios:
                    assert scenario_name not in master._scenarios_included
                    scenario_duals = benders_cut.duals[scenario_name]
                    scenario_ssc = benders_cut.ssc[scenario_name]
                    scenario = scenario_tree.get_scenario(scenario_name)
                    cut_expression += \
                        scenario.probability * \
                        (scenario_ssc + \
                         sum(scenario_duals[variable_id] * \
                             (master_variable[variable_id] - xhat[variable_id]) \
                             for variable_id in xhat))

                cut_expression -= bundle_alpha[i]

        else:
            for scenario in scenario_tree.scenarios:
                if scenario.name in master._scenarios_included:
                    continue
                scenario_name = scenario.name
                scenario_duals = benders_cut.duals[scenario_name]
                scenario_ssc = benders_cut.ssc[scenario_name]
                scenario = scenario_tree.get_scenario(scenario_name)
                cut_expression += \
                    scenario.probability * \
                    (scenario_ssc + \
                     sum(scenario_duals[variable_id] * \
                         (master_variable[variable_id] - xhat[variable_id]) \
                         for variable_id in xhat))

            cut_expression -= master_alpha

        if objective_sense == minimize:
            benders_cuts.append(
                _GeneralConstraintData((None,cut_expression,0.0)))
        else:
            benders_cuts.append(
                _GeneralConstraintData((0.0,cut_expression,None)))

    def extract_master_xhat(self):

        if self.master is None:
            raise RuntimeError("The master problem has not been constructed."
                               "Call the build_master_problem() method to "
                               "construct it.")
        master = self.master
        rootnode = self._manager.scenario_tree.findRootNode()
        master_variable = master.find_component(
            "MASTER_BLEND_VAR_"+str(rootnode.name))
        return dict((variable_id, master_variable[variable_id].value)
                    for variable_id in rootnode._standard_variable_ids
                    if not master_variable[variable_id].stale)

    def solve_master(self):

        if self.master is None:
            raise RuntimeError("The master problem has not been constructed."
                               "Call the build_master_problem() method to "
                               "construct it.")

        common_kwds = {
            'load_solutions':False,
            'tee':self.get_option("master_output_solver_log"),
            'keepfiles':self.get_option("master_keep_solver_files"),
            'symbolic_solver_labels':self.get_option("master_symbolic_solver_labels")}

        if (not self.get_option("master_disable_warmstart")) and \
           (self._master_solver.warm_start_capable()):
            results = self._master_solver.solve(self.master,
                                                warmstart=True,
                                                **common_kwds)
        else:
            results = self._master_solver.solve(self.master,
                                                **common_kwds)

        return results

    def solve(self, **kwds):
        """
        Run the algorithm. If one or both of the keywords max_iterations and
        percent_gap are used, they will override the values on the options
        object used to initialize this class.

        Returns the objective value for the incumbent solution.
        """

        if "max_iterations" not in kwds:
            max_iterations = self.get_option("max_iterations")
        else:
            max_iterations = _domain_positive_integer(
                kwds.pop("max_iterations"))

        if "percent_gap" not in kwds:
            percent_gap = self.get_option("percent_gap")
        else:
            percent_gap = _domain_percent(
                kwds.pop("percent_gap"))

        output_solver_log = kwds.pop("output_solver_log", False)

        assert len(kwds) == 0

        start_time = time.time()

        master = self.master
        master_solver = self._master_solver

        if self.master is None:
            raise RuntimeError("The master problem has not been constructed."
                               "Call the build_master_problem() method to "
                               "construct it.")

        objective_sense = self._manager.objective_sense
        scenario_tree = self._manager.scenario_tree

        def print_dictionary(dictionary):
            #Find longest key
            longest_message = max(len(str(x[0])) for x in dictionary)

            #Find longest dictionary value
            longest_value = max(len(str(x[1])) for x in dictionary)
            for key, value in dictionary:
                print(('{0:<' + str(longest_message) + \
                       '}' '{1:^3}' '{2:<' + str(longest_value) + \
                       '}').format(key,":",value))

#        print("-"*20)
#        print("Problem Statistics")
#        print("-"*20)

        rootnode = scenario_tree.findRootNode()
        num_discrete_firststage = len([variable_id for variable_id
                                       in rootnode._variable_ids \
                                       if rootnode.is_variable_discrete(variable_id)])
        problem_statistics = []
#        problem_statistics.append(("Initial number of cuts in pool"    ,
#                                   len(self.cut_pool)))
#        problem_statistics.append(("Maximum number of iterations"      ,
#                                   max_iterations))
#        problem_statistics.append(("Relative convergence gap threshold",
#                                   percent_gap/100.0))
#        print_dictionary(problem_statistics)


        width_log_table = 100
        if output_solver_log:
            print("")
            print("-"*width_log_table)
            print("%6s %16s %16s %11s  %30s"
                  % ("Iter","Master Bound","Best Incumbent","Gap","Solution Times [s]"))
            print("%6s %16s %16s %11s  %10s %10s %10s %10s"
                  % ("","","","","Master","Sub Min","Sub Max","Cumm"))
            print("-"*width_log_table)

        master_alpha = master.find_component(
            "PYSP_BENDERS_ALPHA_SSC")
        master_bundles_alpha = master.find_component(
            "PYSP_BENDERS_BUNDLE_ALPHA_SSC")
        master_objective = master.find_component(
            "PYSP_BENDERS_OBJECTIVE")

        self.master_bound_history = OrderedDict()
        self.objective_history = OrderedDict()
        self.incumbent_objective = \
            float('inf') if (objective_sense is minimize) else float('-inf')
        self.xhat = None
        self.optimality_gap = float('inf')
        self.iterations = 0

        output_no_gap_warning = True
        for i in xrange(1, max_iterations + 1):

            if (i == 1) and (len(self.cut_pool) == 0):
                # Avoid an unbounded problem. We may still recover
                # an xhat at which to generate a cut, but we can not
                # use the master objective as a lower bound
                master_alpha.fix(0.0)
                master_bundles_alpha.fix(0.0)

            start_time_master = time.time()
            results_master = self.solve_master()
            if len(results_master.solution) == 0:
                raise RuntimeError("Solve failed for master; no solutions generated")
            if results_master.solver.termination_condition != \
               TerminationCondition.optimal:
                logger.warning(
                    "Master solve did not generate an optimal solution:\n"
                    "Solver Status: %s\n"
                    "Solver Termination Condition: %s\n"
                    "Solution Status: %s\n"
                    % (str(results_master.solver.status),
                       str(results_master.solver.termination_condition),
                       str(results_master.solution(0).status)))
            master.solutions.load_from(results_master)
            stop_time_master = time.time()

            if master_alpha.fixed:
                assert i == 1
                assert master_alpha.value == 0.0
                current_master_bound = \
                    float('-inf') if (objective_sense is minimize) else float('inf')
                master_alpha.free()
                master_bundles_alpha.free()
            else:
                current_master_bound = value(master_objective)
                # account for any optimality gap
                solution0 = results_master.solution(0)
                if hasattr(solution0, "gap") and \
                   (solution0.gap is not None) and \
                   (solution0.gap is not undefined):
                    if objective_sense == minimize:
                        current_master_bound -= solution0.gap
                    else:
                        current_master_bound += solution0.gap
                else:
                    if output_no_gap_warning:
                        output_no_gap_warning = False
                        print("WARNING: Solver interface does not "
                              "report an optimality gap. Master bound "
                              "may be invalid.")

            self.master_bound_history[i] = current_master_bound

            new_xhat = self.extract_master_xhat()
            new_cut_info, solve_results = \
                self.generate_cut(new_xhat,
                                  return_solve_results=True)

            # compute the true objective at xhat by
            # replacing the current value of the master cut
            # variable with the true second stage costs of
            # any scenarios involved in the cuts
            self.objective_history[i] = \
                value(master_objective) - value(master_alpha) + \
                sum(scenario.probability * new_cut_info.ssc[scenario.name] \
                    for scenario in scenario_tree.scenarios
                    if scenario.name not in self.master._scenarios_included)

            incumbent_objective_prev = self.incumbent_objective
            best_master_bound = max(self.master_bound_history.values()) if \
                                (objective_sense == minimize) else \
                                min(self.master_bound_history.values())
            self.incumbent_objective = min(self.objective_history.values()) if \
                                       (objective_sense == minimize) else \
                                       max(self.objective_history.values())

            if objective_sense == minimize:
                if self.incumbent_objective < incumbent_objective_prev:
                    self.xhat = new_xhat
            else:
                if self.incumbent_objective > incumbent_objective_prev:
                    self.xhat = new_xhat

            self.optimality_gap = \
                abs(best_master_bound - self.incumbent_objective) / \
                (self.get_option("optimality_gap_epsilon") + \
                 abs(self.incumbent_objective))

            if output_solver_log:
                min_time_sub = min(solve_results.pyomo_solve_time.values())
                max_time_sub = max(solve_results.pyomo_solve_time.values())
                print("%6d %16.4f %16.4f %11.3f%% %10.2f %10.2f %10.2f %10.2f"
                      % (i, current_master_bound, self.incumbent_objective,
                         self.optimality_gap*100, stop_time_master - start_time_master,
                         min_time_sub, max_time_sub, time.time()-start_time))

            # Add the cut even if we exit on this iteration so
            # it ends up in the cut pool (just in case the caller
            # wants to continue the algorithm)
            self.add_cut(new_cut_info)

            # we've completed another iteration
            self.iterations += 1

            # If the optimality gap is below the convergence
            # threshold set by the user, quit the
            # loop. Otherwise, add the new cut to the master
            if self.optimality_gap*100 <= percent_gap:
                if output_solver_log:
                    print("-" * width_log_table)
                    print("Optimality gap threshold reached.")
                break
        else:
            if output_solver_log:
                print("-" * width_log_table)
                print("Maximum number of iterations reached.")

        self.bound = float('-inf') if (objective_sense is minimize) else float('inf')
        if len(self.master_bound_history):
            if objective_sense is minimize:
                self.bound = max(self.master_bound_history.values())
            else:
                self.bound = min(self.master_bound_history.values())

        return self.incumbent_objective

class BendersSolver(SPSolver, PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()
        return BendersAlgorithm._declare_options(options)

    def __init__(self):
        super(BendersSolver, self).__init__(self.register_options())
        self.set_options_to_default()

    def set_options_to_default(self):
        self._options = self.register_options()

    @property
    def options(self):
        return self._options

    @property
    def name(self):
        return "benders"

    def _solve_impl(self,
                    sp,
                    output_solver_log=False):
        with BendersAlgorithm(sp, self._options) as benders:
            benders.initialize_subproblems()
            benders.build_master_problem()
            objective = benders.solve(output_solver_log=output_solver_log)

        results = SPSolverResults()
        results.objective = benders.incumbent_objective
        results.bound = benders.bound
        results.xhat = {sp.scenario_tree.findRootNode().name: benders.xhat}

        return results

def runbenders_register_options(options=None):
    if options is None:
        options = PySPConfigBlock()
    BendersSolver.register_options(options)
    ScenarioTreeManagerFactory.register_options(options)
    safe_register_common_option(options,
                               "verbose")
    safe_register_common_option(options,
                               "disable_gc")
    safe_register_common_option(options,
                               "profile")
    safe_register_common_option(options,
                               "traceback")
    safe_register_common_option(options,
                                "output_solver_log")
    safe_register_common_option(options,
                                "output_scenario_tree_solution")
    return options

#
# Construct a senario tree manager and a BendersAlgorithm
# object to solve it.
#

def runbenders(options):
    """
    Construct a senario tree manager and solve it
    with the Benders solver.
    """
    start_time = time.time()
    with ScenarioTreeManagerFactory(options) as sp:
        sp.initialize()

        print("")
        print("Running Generalized Benders solver for "
              "stochastic programming problems "
              "(i.e., the L-shaped method).")
        benders = BendersSolver()
        benders_options = benders.extract_user_options_to_dict(
            options,
            sparse=True)
        results = benders.solve(
            sp,
            options=benders_options,
            output_solver_log=options.output_solver_log)
        xhat = results.xhat
        del results.xhat
        print("")
        print(results)

        if options.output_scenario_tree_solution:
            print("")
            sp.scenario_tree.snapshotSolutionFromScenarios()
            sp.scenario_tree.pprintSolution()
            sp.scenario_tree.pprintCosts()

    print("")
    print("Total execution time=%.2f seconds"
          % (time.time() - start_time))
    return 0

#
# the main driver routine for the evaluate_xhat script.
#

def main(args=None):
    #
    # Top-level command that executes everything
    #

    #
    # Import plugins
    #
    import pyomo.environ

    #
    # Parse command-line options.
    #
    try:
        options = parse_command_line(
            args,
            runbenders_register_options,
            prog='runbenders',
            description=(
"""Optimize a stochastic program using Generalized Benders
(i.e., the L-shaped method)"""
            ))

    except SystemExit as _exc:
        # the parser throws a system exit if "-h" is specified
        # - catch it to exit gracefully.
        return _exc.code

    return launch_command(runbenders,
                          options,
                          error_label="runbenders: ",
                          disable_gc=options.disable_gc,
                          profile_count=options.profile,
                          traceback=options.traceback)

SPSolverFactory.register_solver("benders", BendersSolver)

if __name__ == "__main__":
    import pyomo.pysp
    sys.exit(main())



import logging
import copy
import math

import pyomo.common.plugin
from pyomo.pysp import phextension
from pyomo.pysp.phutils import indexToString

from pyomo.pysp.plugins.phboundextension import _PHBoundExtensionImpl

logger = logging.getLogger('pyomo.pysp')

class _AdaptiveRhoBase(object):

    def __init__(self):

        self._tol = 1e-5
        self._required_converged_before_decrease = 0
        self._rho_converged_residual_decrease = 1.1
        self._rho_feasible_decrease = 1.25
        self._rho_decrease = 2.0
        self._rho_increase = 2.0
        self._log_rho_norm_convergence_tolerance = 1.0
        self._converged_count = 0
        self._last_adjusted_iter = -1
        self._stop_iter_rho_update = None
        self._prev_avg = None
        self._primal_residual_history = []
        self._dual_residual_history = []
        self._rho_norm_history = []

    def _compute_rho_norm(self, ph):
        rho_norm = 0.0
        for stage in ph._scenario_tree._stages[:-1]:
            for tree_node in stage._tree_nodes:
                node_rho_norm = 0.0
                for variable_id in tree_node._standard_variable_ids:
                    name, index = tree_node._variable_ids[variable_id]
                    # rho is really a node parameter, so we only check
                    # one scenario
                    node_rho_norm += tree_node._scenarios[0].\
                                     _rho[tree_node._name][variable_id]
                rho_norm += node_rho_norm * tree_node._probability
        return rho_norm

    def _snapshot_avg(self, ph):
        self._prev_avg = {}
        for stage in ph._scenario_tree._stages[:-1]:
            for tree_node in stage._tree_nodes:
                self._prev_avg[tree_node._name] = \
                    copy.deepcopy(tree_node._averages)

    def _compute_primal_residual_norm(self, ph):
        primal_resid = {}
        for stage in ph._scenario_tree._stages[:-1]:
            for tree_node in stage._tree_nodes:
                node_primal_resid = primal_resid[tree_node._name] = \
                    dict((variable_id,0.0) for variable_id \
                         in tree_node._standard_variable_ids)
                for variable_id in tree_node._standard_variable_ids:
                    for scenario in tree_node._scenarios:
                        node_primal_resid[variable_id] += \
                            scenario._probability * \
                            (scenario._x[tree_node._name][variable_id] - \
                             tree_node._averages[variable_id])**2
        return primal_resid

    def _compute_dual_residual_norm(self, ph):
        dual_resid = {}
        for stage in ph._scenario_tree._stages[:-1]:
            for tree_node in stage._tree_nodes:
                node_prev_avg = self._prev_avg[tree_node._name]
                dual_resid[tree_node._name] = \
                    dict((variable_id,
                          tree_node._scenarios[0].\
                          _rho[tree_node._name][variable_id]**2 * \
                          (tree_node._averages[variable_id] - \
                           node_prev_avg[variable_id])**2) \
                         for variable_id in tree_node._standard_variable_ids)
        return dual_resid

    def pre_ph_initialization(self,ph):
        pass

    def post_instance_creation(self, ph):
        pass

    def post_ph_initialization(self, ph):
        self._stop_iter_rho_update = int(ph._max_iterations/2)

    def post_iteration_0_solves(self, ph):
        pass

    def post_iteration_0(self, ph):
        pass

    def pre_iteration_k_solves(self, ph):

        if (ph._current_iteration > self._stop_iter_rho_update) and \
           all(not _converger.isConverged(ph) for _converger in ph._convergers):
            return

        converged = any(_converger.isConverged(ph) for _converger in ph._convergers)

        rho_updated = False
        adjust_rho = 0
        if self._prev_avg is None:
            self._snapshot_avg(ph)
        else:
            self._primal_residual_history.append(
                self._compute_primal_residual_norm(ph))
            self._dual_residual_history.append(
                self._compute_dual_residual_norm(ph))
            self._snapshot_avg(ph)
            first_line = ("Updating Rho Values:\n%21s %25s %16s %16s %16s"
                          % ("Action",
                             "Variable",
                             "Primal Residual",
                             "Dual Residual",
                             "New Rho"))
            first = True
            for stage in ph._scenario_tree._stages[:-1]:
                for tree_node in stage._tree_nodes:
                    primal_resid = \
                        math.sqrt(sum(self._primal_residual_history[-1]\
                                      [tree_node._name].values()))
                    dual_resid = \
                        math.sqrt(sum(self._dual_residual_history[-1]\
                                      [tree_node._name].values()))
                    for variable_id in tree_node._standard_variable_ids:
                        name, index = tree_node._variable_ids[variable_id]
                        primal_resid = \
                            math.sqrt(self._primal_residual_history[-1]\
                                      [tree_node._name][variable_id])
                        dual_resid = \
                            math.sqrt(self._dual_residual_history[-1]\
                                      [tree_node._name][variable_id])

                        action = None
                        rho = tree_node._scenarios[0]._rho[tree_node._name][variable_id]
                        if (primal_resid > 10*dual_resid) and (primal_resid > self._tol):
                            rho *= self._rho_increase
                            action = "Increasing"
                        elif ((dual_resid > 10*primal_resid) and (dual_resid > self._tol)):
                            if self._converged_count >= self._required_converged_before_decrease:
                                rho /= self._rho_decrease
                                action = "Decreasing"
                        elif converged:
                            rho /= self._rho_feasible_decrease
                            action = "Feasible, Decreasing"
                        elif (primal_resid < self._tol) and (dual_resid < self._tol):
                            rho /= self._rho_converged_residual_decrease
                            action = "Converged, Decreasing"
                        if action is not None:
                            if first:
                                first = False
                                print(first_line)
                            print("%21s %25s %16g %16g %16g"
                                  % (action, name+indexToString(index),
                                     primal_resid, dual_resid, rho))
                            for scenario in tree_node._scenarios:
                                scenario._rho[tree_node._name][variable_id] = rho

        self._rho_norm_history.append(self._compute_rho_norm(ph))
        if rho_updated:
            print("log(|rho|) = "+repr(math.log(self._rho_norm_history[-1])))

    def post_iteration_k_solves(self, ph):
        pass

    def post_iteration_k(self, ph):
        pass

    def ph_convergence_check(self, ph):

        self._converged_count += 1

        log_rho_norm = math.log(self._compute_rho_norm(ph))
        print("log(|rho|) = "+repr(log_rho_norm))
        if log_rho_norm <= self._log_rho_norm_convergence_tolerance:
            print("Adaptive Rho Convergence Check Passed")
            return True
        print("Adaptive Rho Convergence Check Failed "
              "(requires log(|rho|) < "+
              repr(self._log_rho_norm_convergence_tolerance)+")")
        print("Continuing PH with updated Rho")
        return False

    def post_ph_execution(self, ph):
        pass

class admm(pyomo.common.plugin.SingletonPlugin,
           _AdaptiveRhoBase):

    pyomo.common.plugin.implements(phextension.IPHExtension)

    pyomo.common.plugin.alias("admm")

    def __init__(self):
        _AdaptiveRhoBase.__init__(self)

    def pre_ph_initialization(self,ph):
        _AdaptiveRhoBase.pre_ph_initialization(self, ph)

    def post_instance_creation(self, ph):
        _AdaptiveRhoBase.post_instance_creation(self, ph)

    def post_ph_initialization(self, ph):
        _AdaptiveRhoBase.post_ph_initialization(self, ph)

    def post_iteration_0_solves(self, ph):
        _AdaptiveRhoBase.post_iteration_0_solves(self, ph)

    def post_iteration_0(self, ph):
        _AdaptiveRhoBase.post_iteration_0(self, ph)

    def pre_iteration_k_solves(self, ph):
        _AdaptiveRhoBase.pre_iteration_k_solves(self, ph)

    def post_iteration_k_solves(self, ph):
        _AdaptiveRhoBase.post_iteration_k_solves(self, ph)

    def post_iteration_k(self, ph):
        _AdaptiveRhoBase.post_iteration_k(self, ph)

    def ph_convergence_check(self, ph):
        return _AdaptiveRhoBase.ph_convergence_check(self, ph)

    def post_ph_execution(self, ph):
        _AdaptiveRhoBase.post_ph_execution(self, ph)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import logging
from six import iteritems

from pyomo.common.deprecation import deprecated
from pyomo.core.base import (Transformation,
                             TransformationFactory,
                             Var,
                             Constraint,
                             Objective,
                             Set,
                             minimize,
                             NonNegativeReals,
                             NonPositiveReals,
                             Reals,
                             Block,
                             Model,
                             ConcreteModel)
from pyomo.duality.collect import collect_linear_terms
from pyomo.common.deprecation import deprecated

def load():
    pass

logger = logging.getLogger('pyomo.core')


#
# This transformation creates a new block that
# is the dual of the specified block.  If no block is
# specified, then the entire model is dualized.
# This returns a new Block object.
#
@TransformationFactory.register('duality.linear_dual', doc="Dualize a linear model")
class LinearDual_PyomoTransformation(Transformation):

    @deprecated(
        "Use of the pyomo.duality package is deprecated. There are known bugs "
        "in pyomo.duality, and we do not recommend the use of this code. "
        "Development of dualization capabilities has been shifted to "
        "the Pyomo Adversarial Optimization (PAO) library. Please contact "
        "William Hart for further details (wehart@sandia.gov).",
        version='5.6.2')
    def __init__(self):
        super(LinearDual_PyomoTransformation, self).__init__()

    def _create_using(self, instance, **kwds):
        options = kwds.pop('options', {})
        bname = options.get('block',None)
        #
        # Iterate over the model collecting variable data,
        # until the block is found.
        #
        block = None
        if block is None:
            block = instance
        else:
            for (name, data) in instance.component_map(Block, active=True).items():
                if name == bname:
                    block = instance
        if block is None:
            raise RuntimeError("Missing block: "+bname)
        #
        # Generate the dual
        #
        instance_ = self._dualize(block)

        return instance_

    def _dualize(self, block, unfixed=[]):
        """
        Generate the dual of a block
        """
        #
        # Collect linear terms from the block
        #
        A, b_coef, c_rhs, c_sense, d_sense, vnames, cnames, v_domain = collect_linear_terms(block, unfixed)
        ##print(A)
        ##print(vnames)
        ##print(cnames)
        ##print(list(A.keys()))
        ##print("---")
        ##print(A.keys())
        ##print(c_sense)
        ##print(c_rhs)
        #
        # Construct the block
        #
        if isinstance(block, Model):
            dual = ConcreteModel()
        else:
            dual = Block()
        dual.construct()
        _vars = {}
        def getvar(name, ndx=None):
            v = _vars.get((name,ndx), None)
            if v is None:
                v = Var()
                if ndx is None:
                    v_name = name
                elif type(ndx) is tuple:
                    v_name = "%s[%s]" % (name, ','.join(map(str,ndx)))
                else:
                    v_name = "%s[%s]" % (name, str(ndx))
                setattr(dual, v_name, v)
                _vars[name,ndx] = v
            return v
        #
        # Construct the objective
        #
        if d_sense == minimize:
            dual.o = Objective(expr=sum(- b_coef[name,ndx]*getvar(name,ndx) for name,ndx in b_coef), sense=d_sense)
        else:
            dual.o = Objective(expr=sum(b_coef[name,ndx]*getvar(name,ndx) for name,ndx in b_coef), sense=d_sense)
        #
        # Construct the constraints
        #
        for cname in A:
            for ndx, terms in iteritems(A[cname]):
                expr = 0
                for term in terms:
                    expr += term.coef * getvar(term.var, term.ndx)
                if not (cname, ndx) in c_rhs:
                    c_rhs[cname, ndx] = 0.0
                if c_sense[cname, ndx] == 'e':
                    e = expr - c_rhs[cname,ndx] == 0
                elif c_sense[cname, ndx] == 'l':
                    e = expr - c_rhs[cname,ndx] <= 0
                else:
                    e = expr - c_rhs[cname,ndx] >= 0
                c = Constraint(expr=e)
                if ndx is None:
                    c_name = cname
                elif type(ndx) is tuple:
                    c_name = "%s[%s]" % (cname, ','.join(map(str,ndx)))
                else:
                    c_name = "%s[%s]" % (cname, str(ndx))
                setattr(dual, c_name, c)
            #
            for (name, ndx), domain in iteritems(v_domain):
                v = getvar(name, ndx)
                flag = type(ndx) is tuple and (ndx[-1] == 'lb' or ndx[-1] == 'ub')
                if domain == 1:
                    v.domain = NonNegativeReals
                elif domain == -1:
                    v.domain = NonPositiveReals
                else:
                    # TODO: verify that this case is possible
                    v.domain = Reals

        return dual



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from six.moves import xrange

from pyutilib.misc import Bunch

from pyomo.opt import SolverResults, ProblemSense, SolverStatus, SolutionStatus, TerminationCondition, SolverFactory
from pyomo.opt.blackbox import solver
from pyomo.opt.blackbox import problem


@SolverFactory.register('ps', doc='Pyomo\'s simple pattern search optimizer')
class PatternSearch(solver.COLINSolver):
    """A pattern search optimizer"""

    def __init__(self, **kwds):
        """ Constructor """
        kwds['type'] = "PatternSearch"
        solver.COLINSolver.__init__(self, **kwds)
        self.tmp_point = problem.MixedIntVars()
        #
        # TODO: add a mechanism for managing optiosn
        #
        self.min_function_value = 0.0
        self.max_iterations = 100
        self.min_delta = 1e-5
        self.lower=[]
        self.upper=[]
        self.scale=[]
        self.initial_delta = 1.0
        self.initial_point = None
        self.debug=False

    def reset(self):
        self.delta = self.initial_delta
        if self.problem is not None:
            self.lower = self.problem.real_lower
            self.upper = self.problem.real_upper
        if self.initial_point is None:
            #
            # TODO: set the solver state with this value
            #
            raise ValueError("Pattern Search optimizer must have an initial point specified")
        if not self.feasible(self.initial_point):
            #
            # TODO: set the solver state with this value
            #
            raise ValueError("Pattern Search optimizer initial point must be feasible: "+str(self.initial_point))
        self.x = self.initial_point
        self.scale=[1.0]*len(self.x)
        for i in xrange(0,len(self.x)):
            if type(self.lower[i]) in (float,int) and type(self.upper[i]) in (float,int):
                self.scale[i] = (self.upper[i]-self.lower[i])/10.0

    def feasible(self,point):
        if len(self.lower) == 0:
            return True
        for i in xrange(0,len(self.lower)):
            if self.lower[i] is not None and point[i] < self.lower[i]:
                return False
            if self.upper[i] is not None and point[i] > self.upper[i]:
                return False
        return True

    def evaluate(self,point):
        self.tmp_point.reals = point
        return self.problem.function_value(self.tmp_point)

    def _apply_solver(self):
        """ The routine in a derived class that performs the solve """
        x = self.x
        best = self.evaluate(x)
        ndim=len(x)
        curr_iter=0
        tstatus = TerminationCondition.maxIterations
        tmsg=None
        while curr_iter < self.max_iterations:
            #
            # TODO: use logging here
            #
            if self.debug:      #pragma:nocover
                print("PatternSearch",curr_iter,x,best,self.delta)
            #
            # TODO: add a mechanism to archive the reason for termination
            #
            if best <= self.min_function_value:
                tstatus = TerminationCondition.minFunctionValue
                tmsg = "%s < %s" % (best,self.min_function_value)
                break
            if self.delta <= self.min_delta:
                tstatus = TerminationCondition.minStepLength
                tmsg = "%s < %s" % (self.delta,self.min_delta)
                break
            #
            # Iterate, and keep the best point
            #
            best_new_step=None
            best_new_value=best
            for j in xrange(0,ndim):
                orig=x[j]
                #
                # x[j]-Delta
                #
                x[j] = orig-self.delta*self.scale[j]
                #
                # TODO: use logging here
                #
                if self.debug:          #pragma:nocover
                    print("  Trial Point: ",x)
                if self.lower[j] is None or x[j] >= self.lower[j]:
                    tmp = self.evaluate(x)
                    if tmp < best_new_value:
                        best_new_step=(j,-1)
                        best_new_value=tmp
                #
                # x[j]+Delta
                #
                x[j] = orig+self.delta*self.scale[j]
                #
                # TODO: use logging here
                #
                if self.debug:      #pragma:nocover
                    print("  Trial Point: ",x)
                if self.upper[j] is None or x[j] <= self.upper[j]:
                    tmp = self.evaluate(x)
                    if tmp < best_new_value:
                        best_new_step=(j,1)
                        best_new_value=tmp
                x[j]=orig
            if best_new_value < best:
                #
                # Move to an improving point
                #
                x[best_new_step[0]] = x[best_new_step[0]] + best_new_step[1]*self.delta*self.scale[best_new_step[0]]
                best = best_new_value
                self.delta *= 2.0
            else:
                #
                # Contract about a non-improving point
                #
                self.delta /= 3.0
            curr_iter += 1
        #
        # Archive results in a SolverResults object
        #
        results = SolverResults()
        #
        results.solver.number_of_iterations = curr_iter
        results.solver.final_step_length = self.delta
        results.solver.status = SolverStatus.ok
        results.solver.termination_condition = tstatus
        results.solver.termination_message = tmsg
        #
        results.problem.number_of_objectives=1
        results.problem.number_of_constraints=0
        results.problem.number_of_variables=ndim
        results.problem.sense=ProblemSense.minimize
        results.problem.upper_bound=best
        #
        soln = results.solution.add()
        soln.objective['f'] = {'Value': best}
        soln.status = SolutionStatus.bestSoFar
        for i in xrange(ndim):
            soln.variable["x"+str(i)] = {"Value" : x[i]}
        #
        # Return final results
        #
        self.results = results

        # the pattern search has neither a proper return code or a log:
        return Bunch(rc=None, log=None)



import pyutilib.th as unittest
import pyomo.environ as pe
import pyomo.gdp as gdp
from pyomo.gdp.util import check_model_algebraic
from pyomo.common.log import LoggingIntercept
import logging
from six import StringIO


class TestGDPReclassificationError(unittest.TestCase):
    def test_disjunct_not_in_disjunction(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.d1 = gdp.Disjunct()
        m.d1.c = pe.Constraint(expr=m.x == 1)
        m.d2 = gdp.Disjunct()
        m.d2.c = pe.Constraint(expr=m.x == 0)
        pe.TransformationFactory('gdp.bigm').apply_to(m)
        log = StringIO()
        with LoggingIntercept(log, 'pyomo.gdp', logging.WARNING):
            check_model_algebraic(m)
        self.assertRegexpMatches( log.getvalue(), 
                                  '.*not found in any Disjunctions.*')

    def test_disjunct_not_in_active_disjunction(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.d1 = gdp.Disjunct()
        m.d1.c = pe.Constraint(expr=m.x == 1)
        m.d2 = gdp.Disjunct()
        m.d2.c = pe.Constraint(expr=m.x == 0)
        m.disjunction = gdp.Disjunction(expr=[m.d1, m.d2])
        m.disjunction.deactivate()
        pe.TransformationFactory('gdp.bigm').apply_to(m)
        log = StringIO()
        with LoggingIntercept(log, 'pyomo.gdp', logging.WARNING):
            check_model_algebraic(m)
        self.assertRegexpMatches(log.getvalue(), 
                                 '.*While it participates in a Disjunction, '
                                 'that Disjunction is currently deactivated.*')



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
import pyutilib.th as unittest

import pyomo.environ as aml
from pyomo.pysp.annotations import (locate_annotations,
                                    StageCostAnnotation,
                                    PySP_StageCostAnnotation,
                                    VariableStageAnnotation,
                                    PySP_VariableStageAnnotation,
                                    _ConstraintStageAnnotation,
                                    ConstraintStageAnnotation,
                                    PySP_ConstraintStageAnnotation,
                                    StochasticDataAnnotation,
                                    PySP_StochasticDataAnnotation,
                                    StochasticConstraintBoundsAnnotation,
                                    PySP_StochasticRHSAnnotation,
                                    StochasticConstraintBodyAnnotation,
                                    PySP_StochasticMatrixAnnotation,
                                    StochasticObjectiveAnnotation,
                                    PySP_StochasticObjectiveAnnotation,
                                    StochasticVariableBoundsAnnotation)

class TestAnnotations(unittest.TestCase):

    def test_deprecated(self):
        self.assertIs(StageCostAnnotation,
                      type(PySP_StageCostAnnotation()))
        self.assertIs(VariableStageAnnotation,
                      type(PySP_VariableStageAnnotation()))
        self.assertIs(_ConstraintStageAnnotation,
                      type(PySP_ConstraintStageAnnotation()))
        self.assertIs(_ConstraintStageAnnotation,
                      type(ConstraintStageAnnotation()))
        self.assertIs(StochasticDataAnnotation,
                      type(PySP_StochasticDataAnnotation()))
        self.assertIs(StochasticConstraintBoundsAnnotation,
                      type(PySP_StochasticRHSAnnotation()))
        self.assertIs(StochasticConstraintBodyAnnotation,
                      type(PySP_StochasticMatrixAnnotation()))
        self.assertIs(StochasticObjectiveAnnotation,
                      type(PySP_StochasticObjectiveAnnotation()))

    def _populate_block_with_vars_expressions(self, b):
        b.x = aml.Var()
        b.X1 = aml.Var([1])
        b.X2 = aml.Var([1])
        b.e = aml.Expression()
        b.E1 = aml.Expression([1])
        b.E2 = aml.Expression([1])

    def _populate_block_with_vars(self, b):
        b.x = aml.Var()
        b.X1 = aml.Var([1])
        b.X2 = aml.Var([1])

    def _populate_block_with_constraints(self, b):
        b.x = aml.Var()
        b.c = aml.Constraint(expr= b.x == 1)
        b.C1 = aml.Constraint([1], rule=lambda m, i: m.x == 1)
        b.C2 = aml.Constraint([1], rule=lambda m, i: m.x == 1)
        b.C3 = aml.ConstraintList()
        b.C3.add(b.x == 1)

    def _populate_block_with_objectives(self, b):
        b.x = aml.Var()
        b.o = aml.Objective(expr= b.x + 1)
        b.O1 = aml.Objective([1], rule=lambda m, i: m.x + 1)
        b.O2 = aml.Objective([1], rule=lambda m, i: m.x + 1)

    def _populate_block_with_params(self, b):
        b.p = aml.Param(mutable=True ,initialize=0)
        b.P1 = aml.Param([1], mutable=True, initialize=0)
        b.P2 = aml.Param([1], mutable=True, initialize=0)

    def test_multiple_declarations(self):
        m = aml.ConcreteModel()
        m.x = aml.Var()
        a = StageCostAnnotation()
        a.declare(m, 1)
        a.declare(m.x, 1)
        with self.assertRaises(RuntimeError) as cm:
            a.expand_entries()

    def test_locate_annotations(self):
        m = aml.ConcreteModel()
        m.a = StageCostAnnotation()
        m.b = aml.Block()
        m.b.a = StageCostAnnotation()
        self.assertEqual(locate_annotations(m, StageCostAnnotation),
                         [('a', m.a), ('a', m.b.a)])
        with self.assertRaises(ValueError):
            locate_annotations(m, StageCostAnnotation, max_allowed=1)
        m.b.deactivate()
        self.assertEqual(locate_annotations(m, StageCostAnnotation),
                         [('a', m.a)])
        self.assertEqual(locate_annotations(m, VariableStageAnnotation),
                         [])

    def test_stage_cost(self):
        m = aml.ConcreteModel()
        self._populate_block_with_vars_expressions(m)
        m.b = aml.Block()
        self._populate_block_with_vars_expressions(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_vars_expressions(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_vars_expressions(b))

        a = StageCostAnnotation()
        self.assertEqual(a.default, None)
        self.assertEqual(a.has_declarations, False)
        a.declare(m.x, 1)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.X1, 1)
        a.declare(m.X2[1], 1)
        a.declare(m.e, 1)
        a.declare(m.E1, 1)
        a.declare(m.E2[1], 1)
        with self.assertRaises(TypeError):
            a.declare(m.b, None)
        a.declare(m.b, 1)
        a.declare(m.b_inactive, 1)
        a.declare(m.B, 2)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('x', 1), ('X1[1]', 1), ('X2[1]', 1),
                 ('e', 1), ('E1[1]', 1), ('E2[1]', 1),
                 ('b.x', 1), ('b.X1[1]', 1), ('b.X2[1]', 1),
                 ('b.e', 1), ('b.E1[1]', 1), ('b.E2[1]', 1),
                 ('B[1].x', 2), ('B[1].X1[1]', 2), ('B[1].X2[1]', 2),
                 ('B[1].e', 2), ('B[1].E1[1]', 2), ('B[1].E2[1]', 2)]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('x', 1), ('X1', 1), ('X2[1]', 1),
                 ('e', 1), ('E1', 1), ('E2[1]', 1),
                 ('b.x', 1), ('b.X1', 1), ('b.X2', 1),
                 ('b.e', 1), ('b.E1', 1), ('b.E2', 1),
                 ('B[1].x', 2), ('B[1].X1', 2), ('B[1].X2', 2),
                 ('B[1].e', 2), ('B[1].E1', 2), ('B[1].E2', 2)]))

    def test_variable_stage(self):
        m = aml.ConcreteModel()
        self._populate_block_with_vars_expressions(m)
        m.b = aml.Block()
        self._populate_block_with_vars_expressions(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_vars_expressions(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_vars_expressions(b))

        a = VariableStageAnnotation()
        self.assertEqual(a.default, None)
        self.assertEqual(a.has_declarations, False)
        a.declare(m.x, 1)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.X1, 1)
        a.declare(m.X2[1], 1)
        a.declare(m.e, 1)
        a.declare(m.E1, 1)
        a.declare(m.E2[1], 1)
        with self.assertRaises(TypeError):
            a.declare(m.b, None)
        a.declare(m.b, 1)
        a.declare(m.b_inactive, 1)
        a.declare(m.B, 2, derived=True)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('x', (1,False)), ('X1[1]', (1,False)), ('X2[1]', (1,False)),
                 ('e', (1,False)), ('E1[1]', (1,False)), ('E2[1]', (1,False)),
                 ('b.x', (1,False)), ('b.X1[1]', (1,False)), ('b.X2[1]', (1,False)),
                 ('b.e', (1,False)), ('b.E1[1]', (1,False)), ('b.E2[1]', (1,False)),
                 ('B[1].x', (2,True)), ('B[1].X1[1]', (2,True)), ('B[1].X2[1]', (2,True)),
                 ('B[1].e', (2,True)), ('B[1].E1[1]', (2,True)), ('B[1].E2[1]', (2,True))]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('x', (1,False)), ('X1', (1,False)), ('X2[1]', (1,False)),
                 ('e', (1,False)), ('E1', (1,False)), ('E2[1]', (1,False)),
                 ('b.x', (1,False)), ('b.X1', (1,False)), ('b.X2', (1,False)),
                 ('b.e', (1,False)), ('b.E1', (1,False)), ('b.E2', (1,False)),
                 ('B[1].x', (2,True)), ('B[1].X1', (2,True)), ('B[1].X2', (2,True)),
                 ('B[1].e', (2,True)), ('B[1].E1', (2,True)), ('B[1].E2', (2,True))]))

    def test_constraint_stage(self):
        m = aml.ConcreteModel()
        self._populate_block_with_constraints(m)
        m.b = aml.Block()
        self._populate_block_with_constraints(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_constraints(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_constraints(b))

        a = _ConstraintStageAnnotation()
        self.assertEqual(a.default, None)
        self.assertEqual(a.has_declarations, False)
        a.declare(m.c, 1)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.C1, 1)
        a.declare(m.C2[1], 1)
        a.declare(m.C3, 1)
        with self.assertRaises(TypeError):
            a.declare(m.b, None)
        a.declare(m.b, 1)
        a.declare(m.b_inactive, 1)
        a.declare(m.B, 2)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('c', 1), ('C1[1]', 1), ('C2[1]', 1), ('C3[1]', 1),
                 ('b.c', 1), ('b.C1[1]', 1), ('b.C2[1]', 1), ('b.C3[1]', 1),
                 ('B[1].c', 2), ('B[1].C1[1]', 2), ('B[1].C2[1]', 2), ('B[1].C3[1]', 2)]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('c', 1), ('C1', 1), ('C2[1]', 1), ('C3', 1),
                 ('b.c', 1), ('b.C1', 1), ('b.C2', 1), ('b.C3', 1),
                 ('B[1].c', 2), ('B[1].C1', 2), ('B[1].C2', 2), ('B[1].C3', 2)]))

    def test_stochastic_data(self):
        m = aml.ConcreteModel()
        self._populate_block_with_params(m)
        m.b = aml.Block()
        self._populate_block_with_params(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_params(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_params(b))

        a = StochasticDataAnnotation()
        self.assertEqual(a.default, None)
        self.assertEqual(a.has_declarations, False)
        a.declare(m.p, 1)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.P1, 1)
        a.declare(m.P2[1], 1)
        a.declare(m.b, 1)
        a.declare(m.b_inactive, 1)
        a.declare(m.B, 2)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('p', 1), ('P1[1]', 1), ('P2[1]', 1),
                 ('b.p', 1), ('b.P1[1]', 1), ('b.P2[1]', 1),
                 ('B[1].p', 2), ('B[1].P1[1]', 2), ('B[1].P2[1]', 2)]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('p', 1), ('P1', 1), ('P2[1]', 1),
                 ('b.p', 1), ('b.P1', 1), ('b.P2', 1),
                 ('B[1].p', 2), ('B[1].P1', 2), ('B[1].P2', 2)]))

    def test_constraint_bounds(self):
        m = aml.ConcreteModel()
        self._populate_block_with_constraints(m)
        m.b = aml.Block()
        self._populate_block_with_constraints(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_constraints(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_constraints(b))

        a = StochasticConstraintBoundsAnnotation()
        self.assertEqual(a.default, True)
        self.assertEqual(a.has_declarations, False)
        a.declare(m.c)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.C1)
        a.declare(m.C2[1])
        a.declare(m.C3)
        a.declare(m.b)
        a.declare(m.b_inactive)
        a.declare(m.B, lb=False, ub=True)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('c', True), ('C1[1]', True), ('C2[1]', True), ('C3[1]', True),
                 ('b.c', True), ('b.C1[1]', True), ('b.C2[1]', True), ('b.C3[1]', True),
                 ('B[1].c', (False,True)), ('B[1].C1[1]', (False,True)),
                 ('B[1].C2[1]', (False,True)), ('B[1].C3[1]', (False,True))]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('c', True), ('C1', True), ('C2[1]', True), ('C3', True),
                 ('b.c', True), ('b.C1', True), ('b.C2', True), ('b.C3', True),
                 ('B[1].c', (False,True)), ('B[1].C1', (False,True)),
                 ('B[1].C2', (False,True)), ('B[1].C3', (False,True))]))

    def test_stochastic_constraint_body(self):
        m = aml.ConcreteModel()
        self._populate_block_with_constraints(m)
        m.b = aml.Block()
        self._populate_block_with_constraints(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_constraints(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_constraints(b))

        a = StochasticConstraintBodyAnnotation()
        self.assertEqual(a.default, None)
        self.assertEqual(a.has_declarations, False)
        a.declare(m.c)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.C1)
        a.declare(m.C2[1])
        a.declare(m.C3)
        a.declare(m.b)
        a.declare(m.b_inactive)
        a.declare(m.B, variables=2)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('c', None), ('C1[1]', None), ('C2[1]', None), ('C3[1]', None),
                 ('b.c', None), ('b.C1[1]', None), ('b.C2[1]', None), ('b.C3[1]', None),
                 ('B[1].c', 2), ('B[1].C1[1]', 2), ('B[1].C2[1]', 2), ('B[1].C3[1]', 2)]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('c', None), ('C1', None), ('C2[1]', None), ('C3', None),
                 ('b.c', None), ('b.C1', None), ('b.C2', None), ('b.C3', None),
                 ('B[1].c', 2), ('B[1].C1', 2), ('B[1].C2', 2), ('B[1].C3', 2)]))

    def test_stochastic_objective(self):
        m = aml.ConcreteModel()
        self._populate_block_with_objectives(m)
        m.b = aml.Block()
        self._populate_block_with_objectives(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_objectives(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_objectives(b))

        a = StochasticObjectiveAnnotation()
        self.assertEqual(a.default, (None, True))
        self.assertEqual(a.has_declarations, False)
        a.declare(m.o)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.O1)
        a.declare(m.O2[1])
        a.declare(m.b)
        a.declare(m.b_inactive)
        a.declare(m.B, variables=1, include_constant=False)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('o', (None,True)), ('O1[1]', (None,True)), ('O2[1]', (None,True)),
                 ('b.o', (None,True)), ('b.O1[1]', (None,True)), ('b.O2[1]', (None,True)),
                 ('B[1].o', (1,False)), ('B[1].O1[1]', (1,False)), ('B[1].O2[1]', (1,False))]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('o', (None,True)), ('O1', (None,True)), ('O2[1]', (None,True)),
                 ('b.o', (None,True)), ('b.O1', (None,True)), ('b.O2', (None,True)),
                 ('B[1].o', (1,False)), ('B[1].O1', (1,False)), ('B[1].O2', (1,False))]))

    def test_stochastic_variable_bounds(self):
        m = aml.ConcreteModel()
        self._populate_block_with_vars(m)
        m.b = aml.Block()
        self._populate_block_with_vars(m.b)
        m.b_inactive = aml.Block()
        self._populate_block_with_vars(m.b_inactive)
        m.b_inactive.deactivate()
        m.B = aml.Block([1],
                        rule=lambda b: \
            self._populate_block_with_vars(b))

        a = StochasticVariableBoundsAnnotation()
        self.assertEqual(a.default, (True, True))
        self.assertEqual(a.has_declarations, False)
        a.declare(m.x)
        self.assertEqual(a.has_declarations, True)
        a.declare(m.X1)
        a.declare(m.X2[1])
        a.declare(m.b)
        a.declare(m.b_inactive)
        a.declare(m.B, lb=False, ub=True)
        with self.assertRaises(TypeError):
            a.declare(None, 1)
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries()]),
            set([('x', (True,True)), ('X1[1]', (True,True)), ('X2[1]', (True,True)),
                 ('b.x', (True,True)), ('b.X1[1]', (True,True)), ('b.X2[1]', (True,True)),
                 ('B[1].x', (False,True)), ('B[1].X1[1]', (False,True)), ('B[1].X2[1]', (False,True))]))
        self.assertEqual(
            set([(v[0].name, v[1]) for v in a.expand_entries(expand_containers=False)]),
            set([('x', (True,True)), ('X1', (True,True)), ('X2[1]', (True,True)),
                 ('b.x', (True,True)), ('b.X1', (True,True)), ('b.X2', (True,True)),
                 ('B[1].x', (False,True)), ('B[1].X1', (False,True)), ('B[1].X2', (False,True))]))

if __name__ == "__main__":
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# Test behavior of concrete classes.
#

import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest

import pyomo.opt
from pyomo.environ import *

solvers = pyomo.opt.check_available_solvers('glpk')

@unittest.skipIf(not 'glpk' in solvers, "glpk solver is not available")
class Test(unittest.TestCase):

    def test_blending(self):
        """ The blending example from the PuLP documentation """
        model = ConcreteModel()

        model.x1 = Var(bounds=(0,None), doc="ChickenPercent")
        model.x2 = Var(bounds=(0,None), doc="BeefPercent")

        model.obj = Objective(expr=0.013*model.x1 + 0.008*model.x2, doc="Total Cost of Ingredients per can")

        model.c0 = Constraint(expr=model.x1+model.x2 == 100.0, doc="Percentage Sum")
        model.c1 = Constraint(expr=0.100*model.x1 + 0.200*model.x2 >= 8.0, doc="Protein Requirement")
        model.c2 = Constraint(expr=0.080*model.x1 + 0.100*model.x2 >= 6.0, doc="Fat Requirement")
        model.c3 = Constraint(expr=0.001*model.x1 + 0.005*model.x2 <= 2.0, doc="Fiber Requirement")
        model.c4 = Constraint(expr=0.002*model.x1 + 0.005*model.x2 <= 0.4, doc="Salt Requirement")
        opt = SolverFactory('glpk')
        results = opt.solve(model)
        model.solutions.store_to(results)
        results.write(filename=currdir+"blend.out", format='json')
        self.assertMatchesJsonBaseline(currdir+"blend.out",currdir+"blend.txt", tolerance=1e-2)


if __name__ == "__main__":
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
import pyutilib.th as unittest

from pyomo.environ import ConcreteModel, Block, Var, Reference, Set, Constraint
from pyomo.dae import ContinuousSet
# This inport will have to change when we decide where this should go...
from pyomo.dae.flatten import flatten_dae_components

class TestCategorize(unittest.TestCase):
    def _hashRef(self, ref):
        return tuple(sorted(id(_) for _ in ref.values()))

    def test_flat_model(self):
        m = ConcreteModel()
        m.T = ContinuousSet(bounds=(0,1))
        m.x = Var()
        m.y = Var([1,2])
        m.a = Var(m.T)
        m.b = Var(m.T, [1,2])
        m.c = Var([3,4], m.T)

        regular, time = flatten_dae_components(m, m.T, Var)
        regular_id = set(id(_) for _ in regular)
        self.assertEqual(len(regular), 3)
        self.assertIn(id(m.x), regular_id)
        self.assertIn(id(m.y[1]), regular_id)
        self.assertIn(id(m.y[2]), regular_id)
        # Output for debugging
        #for v in time:
        #    v.pprint()
        #    for _ in v.values():
        #        print"     -> ", _.name
        ref_data = {
            self._hashRef(Reference(m.a[:])),
            self._hashRef(Reference(m.b[:,1])),
            self._hashRef(Reference(m.b[:,2])),
            self._hashRef(Reference(m.c[3,:])),
            self._hashRef(Reference(m.c[4,:])),
        }
        self.assertEqual(len(time), len(ref_data))
        for ref in time:
            self.assertIn(self._hashRef(ref), ref_data)

    def test_1level_model(self):
        m = ConcreteModel()
        m.T = ContinuousSet(bounds=(0,1))
        @m.Block([1,2],m.T)
        def B(b, i, t):
            b.x = Var(list(range(2*i, 2*i+2)))

        regular, time = flatten_dae_components(m, m.T, Var)
        self.assertEqual(len(regular), 0)
        # Output for debugging
        #for v in time:
        #    v.pprint()
        #    for _ in v.values():
        #        print"     -> ", _.name
        ref_data = {
            self._hashRef(Reference(m.B[1,:].x[2])),
            self._hashRef(Reference(m.B[1,:].x[3])),
            self._hashRef(Reference(m.B[2,:].x[4])),
            self._hashRef(Reference(m.B[2,:].x[5])),
        }
        self.assertEqual(len(time), len(ref_data))
        for ref in time:
            self.assertIn(self._hashRef(ref), ref_data)


    def test_2level_model(self):
        m = ConcreteModel()
        m.T = ContinuousSet(bounds=(0,1))
        @m.Block([1,2],m.T)
        def B(b, i, t):
            @b.Block(list(range(2*i, 2*i+2)))
            def bb(bb, j):
                bb.y = Var([10,11])
            b.x = Var(list(range(2*i, 2*i+2)))

        regular, time = flatten_dae_components(m, m.T, Var)
        self.assertEqual(len(regular), 0)
        # Output for debugging
        #for v in time:
        #    v.pprint()
        #    for _ in v.values():
        #        print"     -> ", _.name
        ref_data = {
            self._hashRef(Reference(m.B[1,:].x[2])),
            self._hashRef(Reference(m.B[1,:].x[3])),
            self._hashRef(Reference(m.B[2,:].x[4])),
            self._hashRef(Reference(m.B[2,:].x[5])),
            self._hashRef(Reference(m.B[1,:].bb[2].y[10])),
            self._hashRef(Reference(m.B[1,:].bb[2].y[11])),
            self._hashRef(Reference(m.B[1,:].bb[3].y[10])),
            self._hashRef(Reference(m.B[1,:].bb[3].y[11])),
            self._hashRef(Reference(m.B[2,:].bb[4].y[10])),
            self._hashRef(Reference(m.B[2,:].bb[4].y[11])),
            self._hashRef(Reference(m.B[2,:].bb[5].y[10])),
            self._hashRef(Reference(m.B[2,:].bb[5].y[11])),
        }
        self.assertEqual(len(time), len(ref_data))
        for ref in time:
            self.assertIn(self._hashRef(ref), ref_data)


    def test_2dim_set(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0,1))

        m.v = Var(m.time, [('a',1), ('b',2)])

        scalar, dae = flatten_dae_components(m, m.time, Var)
        self.assertEqual(len(scalar), 0)
        ref_data = {
                self._hashRef(Reference(m.v[:,'a',1])),
                self._hashRef(Reference(m.v[:,'b',2])),
                }
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)

    
    def test_indexed_block(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0,1))
        m.comp = Set(initialize=['a', 'b'])

        def bb_rule(bb, t):
            bb.dae_var = Var()

        def b_rule(b, c):
            b.bb = Block(m.time, rule=bb_rule)

        m.b = Block(m.comp, rule=b_rule)

        scalar, dae = flatten_dae_components(m, m.time, Var)
        self.assertEqual(len(scalar), 0)
        ref_data = {
                self._hashRef(Reference(m.b['a'].bb[:].dae_var)),
                self._hashRef(Reference(m.b['b'].bb[:].dae_var)),
                }
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)


    def test_constraint(self):
        m = ConcreteModel()
        m.time = ContinuousSet(bounds=(0,1))
        m.comp = Set(initialize=['a', 'b'])
        m.v0 = Var()
        m.v1 = Var(m.time)
        m.v2 = Var(m.time, m.comp)
        
        def c0_rule(m):
            return m.v0 == 1
        m.c0 = Constraint(rule=c0_rule)

        def c1_rule(m, t):
            return m.v1[t] == 3
        m.c1 = Constraint(m.time, rule=c1_rule)

        @m.Block(m.time)
        def b(b, t):
            def c2_rule(b, j):
                return b.model().v2[t, j] == 5
            b.c2 = Constraint(m.comp, rule=c2_rule)

        scalar, dae = flatten_dae_components(m, m.time, Constraint)
        hash_scalar = {id(s) for s in scalar}
        self.assertIn(id(m.c0), hash_scalar)

        ref_data = {
                self._hashRef(Reference(m.c1[:])),
                self._hashRef(Reference(m.b[:].c2['a'])),
                self._hashRef(Reference(m.b[:].c2['b'])),
                }
        self.assertEqual(len(dae), len(ref_data))
        for ref in dae:
            self.assertIn(self._hashRef(ref), ref_data)


    # TODO: Add tests for Sets with dimen==None


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright (c) 2014 Sandia Corporation.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  This software is distributed under the BSD License.
#  _________________________________________________________________________

__all__ = ("SPSolverShellCommand",)

import os
import logging
import pprint

import pyutilib.misc

import pyomo.common
from pyomo.pysp.solvers.spsolver import SPSolver

logger = logging.getLogger('pyomo.pysp')

class SPSolverShellCommand(SPSolver):

    def __init__(self, *args, **kwds):
        super(SPSolverShellCommand, self).__init__(*args, **kwds)
        self._executable = None
        self._files = {}

    def _create_tempdir(self, label, *args, **kwds):
        dirname = pyutilib.services.\
                  TempfileManager.create_tempdir(*args, **kwds)
        self._files[label] = dirname
        return dirname

    def _create_tempfile(self, label, *args, **kwds):
        filename = pyutilib.services.\
                   TempfileManager.create_tempfile(*args, **kwds)
        self._files[label] = filename
        return filename

    def _add_tempfile(self, label, filename):
        pyutilib.services.\
            TempfileManager.add_tempfile(filename,
                                         exists=False)
        self._files[label] = filename

    @property
    def executable(self):
        """The executable used by this solver."""
        return self._executable

    @property
    def files(self):
        """A dictionary maintaining the location of various
        solvers files generated during the most recent
        solve. All files will be removed before a solve
        completes unless the keep_solver_files keyword was
        set to True."""
        return self._files

    def set_executable(self, name, validate=True):
        """
        Set the executable for this solver.

        Args:
            name (str): A relative, absolute, or base
                executable name.
            validate (bool): When set to True (default)
                extra checks take place that ensure an
                executable file with that name exists, and
                then 'name' is converted to an absolute
                path. On Windows platforms, a '.exe'
                extension will be appended if necessary when
                validating 'name'. If a file named 'name'
                does not appear to be a relative or absolute
                path, the search will be performed within
                the directories assigned to the PATH
                environment variable.
        """
        if not validate:
            self._executable = name
        else:
            exe = pyomo.common.Executable(name)
            # This is a bit awkward, but we want Executable to re-check
            # the PATH, so we will explicitly call rehash().  In the
            # future, we should move to have the solver directly use the
            # Executable() singleton to manage getting / setting /
            # overriding paths to various executables.  Setting the
            # executable through the Executable() singleton will
            # automatically re-check the PATH.
            exe.rehash()
            exe = exe.path()
            if exe is None:
                raise ValueError(
                    "Failed to set executable for solver %s. File "
                    "with name=%s either does not exist or it is "
                    "not executable. To skip this validation, "
                    "call set_executable with validate=False."
                    % (self.name, name))
            self._executable = exe

    def available(self):
        """Returns whether this solver is available by checking
        if the currently assigned executable exists."""
        exe = self._executable
        try:
            self.set_executable(exe, validate=True)
        except ValueError:
            return False
        else:
            return True
        finally:
            self._executable = exe

    def solve(self, sp, *args, **kwds):
        """
        Solve a stochastic program.

        See the 'solve' method on the base class for
        additional keyword documentation.

        Args:
            sp: The stochastic program to solve.
            keep_solver_files (bool): Retain temporary solver
                input and output files after the solve completes.
            *args: Passed to the derived solver class
                (see the _solve_impl method).
            **kwds: Passed to the derived solver class
                (see the _solve_impl method).

        Returns: A results object with information about the solution.
        """

        self._files.clear()
        assert self.executable is not None

        keep_solver_files = kwds.pop("keep_solver_files", False)
        pyutilib.services.TempfileManager.push()
        try:
            return super(SPSolverShellCommand, self).solve(sp, *args, **kwds)
        finally:
            # cleanup
            pyutilib.services.TempfileManager.pop(
                remove=not keep_solver_files)
            if keep_solver_files:
                logger.info("Retaining the following solver files:\n"
                            +pprint.pformat(self.files))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import sys
import os
import time

from pyomo.core import minimize
from pyomo.pysp.ef_writer_script import ExtensiveFormAlgorithm
from pyomo.pysp.phinit import run_ph
from pyomo.pysp.phutils import (reset_nonconverged_variables,
                                extractVariableNameAndIndex,
                                reset_stage_cost_variables)
from pyomo.pysp.solutionwriter import ISolutionWriterExtension
from pyomo.common.plugin import ExtensionPoint

# Tear the scenario instances off the ef instance when it is no longer required
# so warnings are not generated next time scenarios instances are placed inside
# a new ef instance
def _tear_down_ef(ef_instance, scenario_instances):
   for name in scenario_instances:
      ef_instance.del_component(name)

def solve_ph_code(ph, options):
   import pyomo.environ
   import pyomo.solvers.plugins.smanager.phpyro

   # consolidate the code to solve the problem for the "global" ph object
   # return a solver code (string from the solver if EF, "PH" if PH) and the objective fct val
   SolStatus = None

   ph._preprocess_scenario_instances()

   ObjectiveFctValue = \
      float('inf') if (ph._objective_sense is minimize) else float('-inf')
   SolStatus = None
   if not options.solve_with_ph:
      if options.verbose is True:
         print("Creating the extensive form.")
         print("Time="+time.asctime())

      with ExtensiveFormAlgorithm(ph,
                                  options._ef_options,
                                  options_prefix="ef_") as ef:
         ef.build_ef()
         failure = ef.solve(io_options=\
                            {'output_fixed_variable_bounds':
                             options.write_fixed_variables},
                            exception_on_failure=False)
         if not failure:
            ObjectiveFctValue = ef.objective
         SolStatus = str(ef.solver_status)

      """
      ef = create_ef_instance(ph._scenario_tree,
                              verbose_output=options.verbose)

      if options.verbose:
         print("Time="+time.asctime())
         print("Solving the extensive form.")

      ef_results = solve_ef(ef, options)

      SolStatus = str(ef_results.solver.status) 
      print("SolStatus="+SolStatus)
      if options.verbose is True:
         print("Loading extensive form solution.")
         print("Time="+time.asctime())
      ### If the solution is infeasible, we don't want to load the results
      ### It is up to the caller to decide what to do with non-optimal
      if SolStatus != "infeasible" and SolStatus != "unknown":

         # IMPT: the following method populates the _solution variables on the scenario tree
         #       nodes by forming an average of the corresponding variable values for all
         #       instances particpating in that node. if you don't do this, the scenario tree
         #       doesn't have the solution - and we need this below for variable bounding
         ph._scenario_tree.pullScenarioSolutionsFromInstances()
         ph._scenario_tree.snapshotSolutionFromScenarios()
         if options.verbose is True:
            print("SolStatus="+SolStatus)
            print("Time="+time.asctime())

      _tear_down_ef(ef, ph._instances)
      """
   else:
      if options.verbose:
         print("Solving via Progressive Hedging.")

      run_ph(options, ph)
      ph._scenario_tree.pullScenarioSolutionsFromInstances()
      root_node = ph._scenario_tree._stages[0]._tree_nodes[0]
      ObjectiveFctValue = root_node.computeExpectedNodeCost()
      SolStatus = "PH"
      """
      phretval = ph.solve()
      #print("--------->>>> "+str(phretval))
      SolStatus = "PH"
      if phretval is not None:
         print("Iteration zero solve was not successful for scenario: "+str(phretval))
         if options.verbose is True:
            print("Iteration zero solve was not successful for scenario: "+str(phretval))
         SolStatus = "PHFailAtScen"+str(phretval)

      # TBD - also not sure if PH calls snapshotSolutionFromAverages.
      if options.verbose is True:
         print("Done with PH solve.")

      ##begin copy from phinit
      solution_writer_plugins = ExtensionPoint(ISolutionWriterExtension)
      for plugin in solution_writer_plugins:
         plugin.write(ph._scenario_tree, "ph")

      # store the binding instance, if created, in order to load
      # the solution back into the scenario tree.
      binding_instance = None

      #
      # create the extensive form binding instance, so that we can either write or solve it (if specified).
      #
      if (options.write_ef) or (options.solve_ef):

        # The post-solve plugins may have done more variable
        # fixing. These should be pushed to the instance at this
        # point.
        print("Pushing fixed variable statuses to scenario instances")
        ph._push_all_node_fixed_to_instances()
        total_fixed_discrete_vars, total_fixed_continuous_vars = \
            ph.compute_fixed_variable_counts()
        print("Number of discrete variables fixed "
              "prior to ef creation="
              +str(total_fixed_discrete_vars)+
              " (total="+str(ph._total_discrete_vars)+")")
        print("Number of continuous variables fixed "
              "prior to ef creation="
              +str(total_fixed_continuous_vars)+
              " (total="+str(ph._total_continuous_vars)+")")

        print("Creating extensive form for remainder problem")
        ef_instance_start_time = time.time()
        binding_instance = create_ef_instance(ph._scenario_tree)

        ef_instance_end_time = time.time()
        print("Time to construct extensive form instance=%.2f seconds"
              % (ef_instance_end_time - ef_instance_start_time))

      ph._preprocess_scenario_instances()
      #
      # solve the extensive form and load the solution back into the PH scenario tree.
      # contents from the PH solve will obviously be over-written!
      #
      if options.write_ef:
         output_filename = os.path.expanduser(options.ef_output_file)
         # technically, we don't need the symbol map since we aren't solving it.
         print("Starting to write the extensive form")
         ef_write_start_time = time.time()
         symbol_map = write_ef(binding_instance,
                               output_filename,
                               symbolic_solver_labels=options.symbolic_solver_labels,
                               output_fixed_variable_bounds=options.write_fixed_variables)
         ef_write_end_time = time.time()
         print("Extensive form written to file="+output_filename)
         print("Time to write output file=%.2f seconds"
               % (ef_write_end_time - ef_write_start_time))

      if options.solve_ef:

         # set the value of each non-converged, non-final-stage variable to None -
         # this will avoid infeasible warm-stats.
         reset_nonconverged_variables(ph._scenario_tree, ph._instances)
         reset_stage_cost_variables(ph._scenario_tree, ph._instances)

         ef_results = solve_ef(binding_instance, options)

         print("Storing solution in scenario tree")
         ph._scenario_tree.pullScenarioSolutionsFromInstances()
         ph._scenario_tree.snapshotSolutionFromScenarios()

         ef_solve_end_time = time.time()
         print("Time to solve and load results for the "
               "extensive form=%.2f seconds"
               % (ef_solve_end_time - ef_solve_start_time))

         # print *the* metric of interest.
         print("")
         root_node = ph._scenario_tree._stages[0]._tree_nodes[0]
         print("***********************************************************************************************")
         print(">>>THE EXPECTED SUM OF THE STAGE COST VARIABLES="+str(root_node.computeExpectedNodeCost())+"<<<")
         print("***********************************************************************************************")
         print("")
         print("Extensive form solution:")
         ph._scenario_tree.pprintSolution()
         print("")
         print("Extensive form costs:")
         ph._scenario_tree.pprintCosts()

         solution_writer_plugins = ExtensionPoint(ISolutionWriterExtension)
         for plugin in solution_writer_plugins:
            plugin.write(ph._scenario_tree, "postphef")

      if binding_instance is not None:
         _tear_down_ef(binding_instance, ph._instances)
      """

   print("SolStatus="+str(SolStatus))
   if options.verbose:
      print("Time="+time.asctime())

   ## print "(using PySP Cost vars) ObjectiveFctValue=",ObjectiveFctValue
   return SolStatus, ObjectiveFctValue

###########
def ZeroOneIndexListsforVariable(ph, IndVarName, CCStageNum):
   # return lists across scenarios of the zero value scenarios and one value scenarios
   # for unindexed variable in the ph object for a stage (one based)
   # in this routine we trust that it is binary
   ZerosList = []
   OnesList = []

   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         locval = getattr(instance, IndVarName).value
         #print locval
         if locval < 0.5:
            ZerosList.append(scenario)
         else:
            OnesList.append(scenario)
   return [ZerosList,OnesList]

###########
def PrintanIndexList(IndList):
   # show some useful information about an index list (note: indexes are scenarios)
   print("Zeros:")
   for i in IndList[0]:
      print(i._name+'\n')
   print("Ones:")
   for i in IndList[1]:
      print(i._name+'\n')

###########
def ReturnIndexListNames(IndList):
   ListNames=[[],[]]
   #print "Zeros:"
   for i in IndList[0]:
      ListNames[0].append(i._name)
   #print "Ones:"
   for i in IndList[1]:
      ListNames[1].append(i._name)
   return ListNames

###########
def Set_ParmValue(ph, ParmName, NewVal):
   # set the value of the named parm in all instances (unindexed)
   rootnode = ph._scenario_tree._stages[0]._tree_nodes[0]
   for scenario in rootnode._scenarios:
      instance = ph._instances[scenario._name]
      pm = getattr(instance, ParmName)
      if pm.is_indexed():
         for index in pm:
            pm[index].value = NewVal[index]
      else:
         pm.value = NewVal ##### dlw adds value :(  Jan 2014
      # required for advanced preprocessing that takes
      # place in PySP
      ph._problem_states.\
         user_constraints_updated[scenario._name] = True

###########
def Get_ParmValueOneScenario(ph, scenarioName, ParmName):
   instance = ph._instances[scenarioName]
   pm = getattr(instance, ParmName)
   values = []
   for index in pm:
      values.append(pm[index].value)
   return values

def PurifyIndVar(ph, IndVarName, tolZero=1.e-6):
   rootnode = ph._scenario_tree._stages[0]._tree_nodes[0]
   for scenario in rootnode._scenarios:
      instance = ph._instances[scenario._name]
      pm = getattr(instance, IndVarName)
      for index in pm:
        delta = pm[index].value
        if   abs(delta)    < tolZero: delta = 0
        elif abs(delta-1.) < tolZero: delta = 1
        else:
            print("\n** delta["+str(index)+"," + scenario._name + "] = "+str(delta))
            print("\tnot within " + str(tolZero) + " of 0 or 1 ... causes sys.exit")
            print("**************************************************************\n")
            sys.exit()
        pm[index].value = delta

   return

def FreeAllIndicatorVariables(ph, IndVarName):
   for scenario in ph._scenario_tree._scenarios:
      FreeIndicatorVariableOneScenario(ph,
                                       scenario,
                                       IndVarName)

def FreeIndicatorVariableOneScenario(ph, scenario, IndVarName):
   instance = ph._instances[scenario._name]
   getattr(instance, IndVarName).free()
   # required for advanced preprocessing that takes
   # place in PySP
   ph._problem_states.\
      freed_variables[scenario._name].\
      append((IndVarName, None))

def FixAllIndicatorVariables(ph, IndVarName, value):
   for scenario in ph._scenario_tree._scenarios:
      FixIndicatorVariableOneScenario(ph,
                                      scenario,
                                      IndVarName,
                                      value)

def FixIndicatorVariableOneScenario(ph,
                                    scenario,
                                    IndVarName,
                                    fix_value):

   instance = ph._instances[scenario._name]
   getattr(instance, IndVarName).fix(fix_value)
   # required for advanced preprocessing that takes
   # place in PySP
   ph._problem_states.\
      fixed_variables[scenario._name].\
      append((IndVarName, None))

def FixFromLists(Lists, ph, IndVarName, CCStageNum):
   # fix variables from the lists
   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         fix_value = None
         if scenario in Lists[0]:
            fix_value = 0
         elif scenario in Lists[1]:
            fix_value = 1
         if fix_value is not None:
            # we are assuming no index
            FixIndicatorVariableOneVariable(ph,
                                            scenario,
                                            IndVarName,
                                            fix_value)
   return

###########
def UseListsVariableThenSolve(Lists, ph, IndVarName, CCStageNum):
   # use lists across scenarios of to fix the variable value
   # then solve and return some stuff

   FixFromLists(Lists, ph, IndVarName, CCStageNum)
   LagrangianObj = solve_ph_code(ph)
   b = Compute_ExpectationforVariable(ph, IndVarName, 2)
   print("back from compute exp")

   return LagrangianObj, b

#==============================================
def AlgoExpensiveFlip(TrueForZeros, List, lambdaval, ph, IndVarName, CCStageNum):
   # Flip every delta of every scenario and solve with solve_ph_code(ph)
   # to find scenarios for which we will flip the indicator variable
   # returns list of scenarios sorted by the LagrObj
   if TrueForZeros is True:
      print("ExpensiveFlipAlgo: Fix ZerosToOnes and compute: ")
      value=1
   else:
      print("ExpensiveFlipAlgo: Fix OnesToZeros and compute: ")
      value=0
   D={}
   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         if scenario._name in ReturnIndexListNames(List)[abs(value-1)]:
            FixIndicatorVariableOneScenario(ph,
                                            scenario,
                                            IndVarName,
                                            value)

            LagrObj = solve_ph_code(ph)
            b = Compute_ExpectationforVariable(ph, IndVarName, 2)
            z = LagrObj+(b*lambdaval)
            D[z]=[scenario,b]
            print("ObjVal w/ "+str(scenario._name)+
                  " delta flipped to "+str(value)+
                  " (LagrObj,b,z): "+str(LagrObj)+
                  " "+str(b)+" "+str(z))
            FixIndicatorVariableOneScenario(ph,
                                            scenario,
                                            IndVarName,
                                            abs(value-1))

   Dsort = []
   for key in sorted(D):
      Dsort.append(D[key])
   print("---SmallestObjValue:--->  "
         +str(sorted(D)[0])+"  with (LagrObj,b) = "
         +str(sorted(D)[0]-(lambdaval*Dsort[0][1]))+" "
         +str(Dsort[0][1])+"  at  "+str(Dsort[0][0]._name))
   if TrueForZeros is True:
      print("Back from ExpensiveFlipAlgo:Fix ZerosToOnes")
   else:
      print("Back from ExpensiveFlipAlgo:Fix OnesToZeros")
   return Dsort,sorted(D)[0],Dsort[0][1]

def Compute_ExpectationforVariable(ph, IndVarName, CCname, CCStageNum):
   rootnode = ph._scenario_tree._stages[0]._tree_nodes[0]
   ReferenceInstance = ph._instances[rootnode._scenarios[0]._name]  # arbitrary scenario
   CC = getattr(ReferenceInstance,CCname)

   SumSoFar = {}
   for cc in CC: SumSoFar[cc] = 0.0
   node_probability = 0.0
   stage = ph._scenario_tree._stages[CCStageNum-1]
   for tree_node in stage._tree_nodes:
      for scenario in tree_node._scenarios:
         instance = ph._instances[scenario._name]
         node_probability += scenario._probability
         for cc in CC:
            deltaValue = getattr(instance,IndVarName)[cc].value
            SumSoFar[cc] += scenario._probability * deltaValue
   for cc in CC:
      SumSoFar[cc] = SumSoFar[cc] / node_probability
   return SumSoFar

def GetPenaltyCost(ph, IndVarName, multName):
# E[lambda*delta] contribution to cost
   rootnode = ph._scenario_tree._stages[0]._tree_nodes[0]
   ReferenceInstance = ph._instances[rootnode._scenarios[0]._name]  # arbitrary scenario
   CC = ReferenceInstance.ChanceConstraints
   cost = 0.
   for scenario in rootnode._scenarios:
      instance = ph._instances[scenario._name]
      delta = getattr(instance, IndVarName)
      mult  = getattr(instance,multName)
      for index in delta:
        cost = cost + scenario._probability*delta[index].value*mult[index].value
   return cost



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import pyutilib.th as unittest

from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_register_common_option)
import six
from six import StringIO

class TestPySPConfigValue(unittest.TestCase):

    def test_init(self):
        PySPConfigValue(
            "default",
            domain=str,
            description="a description",
            doc=None,
            visibility=0)

class TestPySPConfigBlock(unittest.TestCase):

    def test_init(self):
        PySPConfigBlock()

    def test_declared(self):
        b = PySPConfigBlock()
        safe_register_common_option(b, "verbose")
        b.display()
        b.display()
        out = StringIO()
        b.display(ostream=out)
        self.assertEqual(out.getvalue(),
                         "verbose: false\n")
        self.assertEqual(b.check_usage(), True)
        self.assertEqual(b.check_usage(error=False), True)
        b.verbose = True
        out = StringIO()
        b.display(ostream=out)
        self.assertEqual(out.getvalue(),
                         "verbose: true\n")
        with self.assertRaises(ValueError):
            b.check_usage()
        with self.assertRaises(ValueError):
            b.check_usage()
        self.assertEqual(b.check_usage(error=False), False)
        b.verbose
        self.assertEqual(b.check_usage(), True)
        self.assertEqual(b.check_usage(error=False), True)
        verbose_about = \
"""PySPConfigValue: verbose
  -    type: <%s 'bool'>
  - default: False
  -    doc: Generate verbose output for both initialization and
            execution.""" % ('class' if six.PY3 else 'type')
        self.assertEqual(b.about("verbose"),
                         verbose_about)

    def test_implicit(self):
        b = PySPConfigBlock()
        b._implicit_declaration = True
        b.name_a = 1
        b.nameb = 2
        b.display()
        out = StringIO()
        b.display(ostream=out)
        self.assertEqual(out.getvalue(),
                         "name_a: 1\nnameb: 2\n")
        with self.assertRaises(ValueError):
            b.check_usage()
        with self.assertRaises(ValueError):
            b.check_usage()
        self.assertEqual(b.check_usage(error=False), False)
        b.name_a
        b.nameb
        self.assertEqual(b.check_usage(), True)
        self.assertEqual(b.check_usage(error=False), True)
        name_a_about = \
"""ConfigValue: name_a
  -    type: None
  - default: 1
  -    doc: None"""
        self.assertEqual(b.about("name_a"),
                         name_a_about)

if __name__ == "__main__":
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from pyomo.common.plugin import *
from pyomo.pysp import solutionwriter
from pyomo.pysp.scenariotree import *
from pyomo.pysp.plugins.phhistoryextension \
    import extract_scenario_tree_structure, \
           extract_scenario_solutions, \
           extract_node_solutions

import json

class JSONSolutionWriter(SingletonPlugin):

    implements (solutionwriter.ISolutionWriterExtension)

    def write(self, scenario_tree, output_file_prefix):

        if not isinstance(scenario_tree, ScenarioTree):
            raise RuntimeError("JSONSolutionWriter write method expects ScenarioTree object - type of supplied object="+str(type(scenario_tree)))

        include_ph_objective_parameters = None
        include_variable_statistics = None
        if output_file_prefix == 'ph':
            include_ph_objective_parameters = True
            include_variable_statistics = True
        elif output_file_prefix == 'postphef':
            include_ph_objective_parameters = False
            include_variable_statistics = True
        elif output_file_prefix == 'ef':
            include_ph_objective_parameters = False
            include_variable_statistics = False
        else:
            raise ValueError("JSONSolutionWriter requires an output prefix of 'ef', 'ph', or 'postphef' "
                             "to indicate whether ph specific parameter values should be extracted "
                             "from the solution")
        
        output_filename = output_file_prefix+"_solution.json"
        results = {}
        results['scenario tree'] = extract_scenario_tree_structure(scenario_tree)
        results['scenario solutions'] \
            = extract_scenario_solutions(scenario_tree,
                                         include_ph_objective_parameters=include_ph_objective_parameters)
        results['node solutions'] \
            = extract_node_solutions(scenario_tree,
                                     include_ph_objective_parameters=include_ph_objective_parameters,
                                     include_variable_statistics=include_variable_statistics)

        with open(output_filename,'w') as f:
            json.dump(results,f,indent=2)
        print("Scenario tree solution written to file="+output_filename)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from pyomo.core.expr.numvalue import \
   RegisterNumericType, \
   RegisterIntegerType, \
   RegisterBooleanType

try:
    import numpy
    _has_numpy = True
except:     #pragma:nocover
    _has_numpy = False

#
# Collect NumPy Types
#
# The full list of available types is dependent on numpy version and
# the compiler used to build numpy. The following list of types was
# taken from the NumPy v1.8 Manual at scipy.org
#

# Boolean
numpy_bool_names = []
numpy_bool_names.append('bool_')
numpy_bool = []
if _has_numpy:
    for _type_name in numpy_bool_names:
        try:
            _type = getattr(numpy,_type_name)
            numpy_bool.append(_type)
        except:     #pragma:nocover
            pass

# Integers
numpy_int_names = []
numpy_int_names.append('int_')
numpy_int_names.append('intc')
numpy_int_names.append('intp')
numpy_int_names.append('int8')
numpy_int_names.append('int16')
numpy_int_names.append('int32')
numpy_int_names.append('int64')
numpy_int_names.append('uint8')
numpy_int_names.append('uint16')
numpy_int_names.append('uint32')
numpy_int_names.append('uint64')
numpy_int = []
if _has_numpy:
    for _type_name in numpy_int_names:
        try:
            _type = getattr(numpy,_type_name)
            numpy_int.append(_type)
        except:     #pragma:nocover
            pass

# Reals
numpy_float_names = []
numpy_float_names.append('float_')
numpy_float_names.append('float16')
numpy_float_names.append('float32')
numpy_float_names.append('float64')
numpy_float = []
if _has_numpy:
    for _type_name in numpy_float_names:
        try:
            _type = getattr(numpy,_type_name)
            numpy_float.append(_type)
        except:     #pragma:nocover
            pass

# Complex
numpy_complex_names = []
numpy_complex_names.append('complex_')
numpy_complex_names.append('complex64')
numpy_complex_names.append('complex128')
numpy_complex = []
if _has_numpy:
    for _type_name in numpy_complex_names:
        try:
            _type = getattr(numpy,_type_name)
            numpy_complex.append(_type)
        except:     #pragma:nocover
            pass


#
# Register NumPy Types
#

# Register NumPy boolean types as Boolean
for _type in numpy_bool:
    RegisterBooleanType(_type)

# Register NumPy integer types as Integer (this will also update
# Numeric/Reals) and as Boolean
for _type in numpy_int:
    RegisterIntegerType(_type)
    RegisterBooleanType(_type)

# Register NumPy float types as Numeric and as Boolean
for _type in numpy_float:
    RegisterNumericType(_type)
    RegisterBooleanType(_type)

# Note: If complex types are to be registered it will need to be
#       with a different registration function than
#       RegisterNumericType because this updates the type set used
#       by Reals.

#for _type in numpy_complex:
#    pass




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
Tight bounds should be correlated with effective W values; that is almost a tautology. 

A closely related measure of quality is generated by the konvw extension that produces wonly.ssv showing the convergence of each variable when the proximal term is *not* used.  A truly good W should produce convergence without the proximal term (which is zero at convergence).
"""

import logging
from math import fabs

import pyomo.common.plugin
from pyomo.pysp import phextension
from pyomo.pysp.generators import scenario_tree_node_variables_generator_noinstances
from pyomo.pysp.phutils import indexToString

from operator import itemgetter

from pyomo.pysp.plugins.phboundextension import _PHBoundBase

logger = logging.getLogger('pyomo.pysp')


class phweightinspectextension(pyomo.common.plugin.SingletonPlugin, _PHBoundBase):

    pyomo.common.plugin.implements(phextension.IPHExtension)

    pyomo.common.plugin.alias("phweightinspectextension")

    def __init__(self):

        _PHBoundBase.__init__(self)
        self.wonly_file = "wonly.ssv"
        print ("konvw is creating W-only convergence file="+self.wonly_file)
        ofile = open(self.wonly_file, 'w')
        ofile.close()
        self._valid_weights_objective_relative_tolerance = 0.01

    def _inspect_variable_convergence(self, ph, ph_iter):

        # collect termdiff by node and variable so we can
        # report and possibly sort
        term_diff = dict((tree_node._name, {}) \
                         for stage in ph._scenario_tree._stages[:-1]
                         for tree_node in stage._tree_nodes)

        # Track these for reporting purposes
        node_fixed_cnt = dict((tree_node._name, 0) \
                              for stage in ph._scenario_tree._stages[:-1]
                              for tree_node in stage._tree_nodes)
        total_fixed_cnt = 0

        for stage, tree_node, variable_id, variable_values, is_fixed, is_stale \
            in scenario_tree_node_variables_generator_noinstances(
                ph._scenario_tree,
                includeDerivedVariables=False,
                includeLastStage=False):

            if is_fixed:
                node_fixed_cnt[tree_node._name] += 1
                total_fixed_cnt += 1

            # Depending on preprocessing options, stale may indicate
            # fixed or unused in the model, either way we can skip it
            if (not is_stale):

                var_node_avg = 0.0
                for var_value, scenario_probability in variable_values:
                    var_node_avg += scenario_probability * var_value

                var_term_diff = 0.0
                for var_value, scenario_probability in variable_values:
                    var_term_diff += \
                        scenario_probability * \
                        fabs(var_value - var_node_avg)

                term_diff[tree_node._name][variable_id] = var_term_diff


        # Print individual variable term diffs by node
        # and sorted highest to lowest
        # skip the leaf stage
        ofile = open(self.wonly_file, 'a')
        for stage in ph._scenario_tree._stages[:-1]:

            for tree_node in stage._tree_nodes:

                for variable_id, var_term_diff in sorted(term_diff[tree_node._name].items(),
                                                         key=itemgetter(1),
                                                         reverse=True):
                    variable_name, index = tree_node._variable_ids[variable_id]
                    ofile.write(str(ph_iter)+"; "+tree_node._name+"; "+variable_name+indexToString(index)+"; "+str(var_term_diff)+'\n')
        ofile.close()

    def _iteration_k_solves(self,ph, storage_key):

        # Extract a candidate solution to compute an upper bound
        candidate_sol = self.ExtractInternalNodeSolutionsWithDiscreteRounding(ph)
        # ** Code uses the values stored in the scenario solutions
        #    to perform a weighted vote in the case of discrete
        #    variables, so it is important that we execute this
        #    before perform any new subproblem solves.
        #candidate_sol = self.ExtractInternalNodeSolutionsWithDiscreteVoting(ph)

        # Caching the current set of ph solutions so we can restore
        # the original results. We modify the scenarios and re-solve -
        # which messes up the warm-start, which can seriously impact
        # the performance of PH. plus, we don't want lower bounding to
        # impact the primal PH in any way - it should be free of any
        # side effects.
        self.CachePHSolution(ph)

        #
        #
        #
        #
        # **** NOTE WE ARE LEAVING VARIABLES FIXED FOR NOW ****
        #            THIS IS STILL EXPERIMENTAL
        #
        #
        #

        # Assuming the weight terms are already active but proximal
        # terms need to be deactivated deactivate all proximal terms
        # and activate all weight terms
        self.DeactivatePHObjectiveProximalTerms(ph)

        # Weights have not been pushed to instance parameters (or
        # transmitted to the phsolverservers) at this point
        ph._push_w_to_instances()

        ph.solve_subproblems(warmstart=not ph._disable_warmstarts)

        print("Successfully completed PH weight inspection extension "
              "iteration %s solves\n"
              "- solution statistics:\n" % (storage_key))
        if ph._scenario_tree.contains_bundles():
            ph.report_bundle_objectives()
        ph.report_scenario_objectives()

        weight_only_ph_objective = {}
        for scenario in ph._scenario_tree._scenarios:
            weight_only_ph_objective[scenario._name] = scenario._objective

        #
        # Fix variables to XBAR (weighted vote for discrete) and solve again
        #
        self.FixScenarioTreeVariables(ph, candidate_sol)

        ph.solve_subproblems(warmstart=not ph._disable_warmstarts)

        print("Successfully completed PH weight inspection extension "
              "iteration %s solves (FIXED TO XBAR)\n"
              "- solution statistics:\n" % (storage_key))
        if ph._scenario_tree.contains_bundles():
            ph.report_bundle_objectives()
        ph.report_scenario_objectives()

        fixed_xbar_ph_objective = {}
        for scenario in ph._scenario_tree._scenarios:
            fixed_xbar_ph_objective[scenario._name] = scenario._objective

        print("")
        print("Weight Inspection Results: (Using relative tolerance: "+repr(self._valid_weights_objective_relative_tolerance)+")")
        failure = False
        for scenario in ph._scenario_tree._scenarios:
            error = fabs(fixed_xbar_ph_objective[scenario._name] - \
                         weight_only_ph_objective[scenario._name]) / fabs(fixed_xbar_ph_objective[scenario._name])
            if error > self._valid_weights_objective_relative_tolerance:
                print("\t"+str(scenario._name)+": FAIL (relative error: "+repr(error)+")")
                failure = True
            else:
                print("\t"+str(scenario._name)+": OKAY (relative error: "+repr(error)+")")
        if failure:
            print("******************************")
            print("  Weight Inspection Failed!   ")
            print("******************************")
        else:
            print("")
            print("Weight Inspection Okay")
            print("")
        print("")
        self._inspect_variable_convergence(ph, storage_key)

        # Restore ph to its state prior to entering this method
        # (e.g., fixed variables, scenario solutions, proximal terms)
        self.RestorePH(ph)

    ############ Begin Callback Functions ##############

    def reset(self, ph):
        self.__init__()

    def pre_ph_initialization(self,ph):
        """
        Called before PH initialization.
        """
        pass

    def post_instance_creation(self, ph):
        """
        Called after PH initialization has created the scenario
        instances, but before any PH-related
        weights/variables/parameters/etc are defined!
        """
        pass

    def post_ph_initialization(self, ph):
        """
        Called after PH initialization
        """

        if ph._verbose:
            print("Invoking post initialization callback in phboundextension")

    def post_iteration_0_solves(self, ph):
        """
        Called after the iteration 0 solves
        """
        pass

    def post_iteration_0(self, ph):
        """
        Called after the iteration 0 solves, averages computation, and weight computation
        """
        pass

    def pre_iteration_k_solves(self, ph):
        """
        Called immediately before the iteration k solves
        """
        pass

    def post_iteration_k_solves(self, ph):
        """
        Called after the iteration k solves!
        """
        pass

    def post_iteration_k(self, ph):
        """
        Called after the iteration k is finished, after weights have been updated!
        """
        if ph._verbose:
            print("Invoking post iteration k callback in phboundextension")

        if ph._converger.isConverged(ph):
            ph_iter = ph._current_iteration
            self._iteration_k_solves(ph, ph_iter)

    def post_ph_execution(self, ph):
        """
        Called after PH has terminated!
        """
        pass




# -*- coding: utf-8 -*- 
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# Unit Tests for pyomo.base.misc
#
from six import StringIO, iteritems

import pyutilib.th as unittest

from pyomo.core.base.misc import tabular_writer

class TestTabularWriter(unittest.TestCase):
    def test_unicode_table(self):
        # Test that an embedded unicode character does not foul up the
        # table alignment
        os = StringIO()
        data = {1: ("a", 1), (2,3): ("∧", 2)}
        tabular_writer(os, "", iteritems(data), ["s", "val"], lambda k,v: v)
        ref = u"""
Key    : s : val
     1 : a :   1
(2, 3) : ∧ :   2
"""
        self.assertEqual(ref.strip(), os.getvalue().strip())

if __name__ == "__main__":
    unittest.main()
    
